/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/*          EXIFPROBE - TIFF/JPEG/EXIF image file probe               */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* Copyright (C) 2002 by Duane H. Hesser. All rights reserved.        */
/*                                                                    */
/* See the file LICENSE.EXIFPROBE for terms of use.                   */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#ifndef lint
static char *ModuleId = "@(#) $Id: maker_asahi.c,v 1.4 2005/01/27 17:36:23 alex Exp $";
#endif

/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* Asahi/Pentax camera maker-specific routines                        */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* This information is from ab image by Johannes Tschebisch at:       */
/* http://www.jojotsch.de/downloads/jojothumb/beispiele/              */
/*                             html_exif/bilder/bilder.html           */
/*                                                                    */

/* New info (Jan 05) from:                                            */
/* http://www.ozhiker.com/electronics/pjmt/jpeg_info/pentax_mn.html   */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "extern.h"
#include "misc.h"
#include "maker.h"

/* Find the identifying number assigned to known Asahi/Pentax camera  */
/* models. This number is used to dispatch print and interpret        */
/* routines approopriate to the current image.                        */

int
asahi_model_number(char *model,char *software)
{
    int number = GENERIC;

    if(model)
    {
        switch(*model)
        {
                case 'P':
                    if(strncmp(model,"PENTAX Optio430RS",17) == 0)
                        number = ASAHI_PENTAX430RS;
                    else if(strncmp(model,"PENTAX Optio330RS",17) == 0)
                        number = ASAHI_PENTAX330RS;
                    else if(strncmp(model,"PENTAX Optio 330",16) == 0)
                        number = ASAHI_PENTAX330;
                    else if(strncmp(model,"PENTAX Optio 430",16) == 0)
                        number = ASAHI_PENTAX430;
                break;
            default:
                break;
        }
    }
    return(number);
}


/* Dispatch a print routine for direct values in Asahi/Pentax         */
/* cameras, based upon model                                          */

void
print_asahi_makervalue(struct ifd_entry *entry_ptr,int make,int model,
                    char *prefix)
{

    if(entry_ptr && (PRINT_VALUE))
    {
        switch(model)
        {
        case ASAHI_PENTAX330RS:
        case ASAHI_PENTAX430RS:
                print_asahi1_makervalue(entry_ptr,make,model,prefix);
                asahi2_interpret_value(entry_ptr);
            break;
        case ASAHI_PENTAX330:
        case ASAHI_PENTAX430:
            case GENERIC:
            default:
                print_asahi1_makervalue(entry_ptr,make,model,prefix);
                asahi1_interpret_value(entry_ptr);
                break;
        }
    }
}

/* Model-specific print routine for Asahi/Pentax cameras. This        */
/* routine is responsible for picking off any direct entry tags which */
/* are peculiar and will not be handled properly by print_value()     */
/* (usually UNDEFINED values which fit in the 4-byte entry value). If */
/* there are no such entries, this function simply calls              */
/* print_value().                                                     */

void
print_asahi1_makervalue(struct ifd_entry *entry_ptr,int make,int model,
                    char *prefix)
{
    if(entry_ptr && (PRINT_VALUE))
    {
        switch(entry_ptr->tag) 
        {
            default:
                print_value(entry_ptr,prefix);
                break;
        }
    }
}


/* Dispatch a routine to decode and print offset values for Asahi/Pentax     */
/* cameras.                                                           */

void
print_asahi_offset_makervalue(FILE *inptr,unsigned short byteorder,
    struct ifd_entry *entry_ptr,unsigned long offset_ref,char *prefix,
    int indent,int addrwidth,int make,int model,int at_offset)
{

    if(entry_ptr && (PRINT_VALUE))
    {
        switch(model)
        {
            /* currently only one routine needed                          */
            default:
                asahi1_offset_makervalue(inptr,byteorder,entry_ptr,
                                            offset_ref,prefix,indent,
                                            addrwidth,make,model,at_offset);
                break;
        }
    }
}

/* Model-specific routine to print UNDEFINED values found at offsets  */
/* in Asahi/Pentax makernotes.                                        */

void
asahi1_offset_makervalue(FILE *inptr,unsigned short byteorder,
    struct ifd_entry *entry_ptr,unsigned long offset_ref,char *prefix,
    int indent,int addrwidth,int make,int model,int at_offset)
{
    static unsigned long jpegthumbnaillength = 0UL;
    unsigned long value_offset;
    unsigned long dumplength;
    unsigned short marker;
    unsigned short count;
    char *nameoftag;

    if(inptr && entry_ptr)
    {
        nameoftag = maker_tagname(entry_ptr->tag,make,model); 
        value_offset = offset_ref + entry_ptr->value;
        count = entry_ptr->count;

        switch(entry_ptr->tag)
        {
            case 0x0004:    /* JpegThumbnailOffset                    */
                /* This model provides the 0x2000 tag as well as the  */
                /* 0x0003/0x0004 tags. Both entries point to the same */
                /* thumbnail, so just print values for length and     */
                /* offset tags, and display the thumbnail on the      */
                /* (self-contained) 0x2000 tag                        */
                if(!(PRINT_OFFSET))
                {
                    if(PRINT_BOTH_VALUE)
                        printf("@%#lx=%lu",value_offset,value_offset);
                    else if(PRINT_HEX_VALUE)
                        printf("@%#lx",value_offset);
                    else if(PRINT_DEC_VALUE)
                        printf("@%lu",value_offset);
                }
                break;
            case 0x2000: /* Jpeg Thumbnail as UNDEFINED section       */
                if(at_offset)
                {
                    print_tag_address(VALUE_AT_OFFSET,value_offset,indent,prefix,addrwidth);
                    extraindent(SUBINDENT);
                    printf("Start of JPEG Thumbnail");
                }
                if(!(PRINT_OFFSET))
                {
                    if(PRINT_BOTH_VALUE)
                        printf("@%#lx=%lu",value_offset,value_offset);
                    else if(PRINT_HEX_VALUE)
                        printf("@%#lx",value_offset);
                    else if(PRINT_DEC_VALUE)
                        printf("@%lu",value_offset);
                }
                printf(" length %ld",count);
                marker = read_imagemarker(inptr,value_offset);
                if(PRINT_SEGMENT)
                {
                    putchar('\n');
                    (void)process_jpeg_segments(inptr,value_offset,
                                                marker,count,
                                                indent+(SUBINDENT * 2),addrwidth);
                    print_tag_address(SEGMENT,value_offset + count - 1,
                                                indent,"-",addrwidth);
                    extraindent(SUBINDENT);
                    printf("End of JPEG Thumbnail");
                }

                /* make certain we're at the end                      */
                // ###%%% perhaps we shouldn't do this when the thumbnail
                // is located past the end of the makernote?
                fseek(inptr,value_offset + count,0);
                break;
            case 0x0e00: /* PrintIM (Epson Print Image matching)      */
                if(!(PRINT_OFFSET))
                {
                    if(PRINT_BOTH_VALUE)
                        printf("@%#lx=%lu\n",value_offset,value_offset);
                    else if(PRINT_HEX_VALUE)
                        printf("@%#lx\n",value_offset);
                    else if(PRINT_DEC_VALUE)
                        printf("@%lu\n",value_offset);
                }
                process_pim(inptr,byteorder,entry_ptr->value,offset_ref,
                    entry_ptr->count,nameoftag,"Exif_Maker_",prefix,indent,addrwidth);
                break;
            default:
                print_generic_offset_makervalue(inptr,byteorder,entry_ptr,
                                        offset_ref,prefix,indent,
                                        addrwidth,make,model,at_offset);
                break;
        }
    }
}


/* Asahi/Pentax-specific tagnames for makernotes.                     */

char *
maker_asahi_tagname(unsigned short tag,int model)
{
    char *tagname = (char *)0;

    /* Check specific models first                                    */
    switch(model)
    {
        case ASAHI_PENTAX330RS:
        case ASAHI_PENTAX430RS:
            tagname = maker_asahi2_tagname(tag,model);
            break;
        case ASAHI_PENTAX330:
        case ASAHI_PENTAX430:
            tagname = maker_asahi1_tagname(tag,model);
            break;
        /* Add special-case models here                               */
        default:
            tagname = maker_asahi1_tagname(tag,model);
            break;
    }

    /* If no model-specific tag is found, use a "generic"             */
    /* Asahi/Pentax tag                                               */
    if(tagname == NULL)
    {
        switch(tag)
        {
            case 0x0e00: tagname = "PrintIM"; break;
            default: break;
        }
    }
    return(tagname);
}


char *
maker_asahi1_tagname(unsigned short tag,int model)
{
    char *tagname = (char *)0;

    switch(tag)
    {
        case 0x0001: tagname = "CaptureMode"; break;
        case 0x0002: tagname = "Quality"; break;
        case 0x0003: tagname = "FocusMode"; break;
        case 0x0004: tagname = "FlashMode"; break;
        case 0x0007: tagname = "WhiteBalance"; break;
        case 0x000a: tagname = "DigitalZoom"; break;
        case 0x000b: tagname = "Sharpness"; break;
        case 0x000c: tagname = "Contrast"; break;
        case 0x000d: tagname = "Saturation"; break;
        case 0x0014: tagname = "ISOSpeedSetting"; break;
        case 0x0017: tagname = "Color"; break;
        case 0x1000: tagname = "TimeZone"; break;
        case 0x1001: tagname = "DaylightSavings"; break;
    }
    return(tagname);
}

char *
maker_asahi2_tagname(unsigned short tag,int model)
{
    char *tagname = (char *)0;

    switch(tag)
    {
        case 0x0002: tagname = "JpegThumbnailDimensions"; break;
        case 0x0003: tagname = "JpegThumbnailLength"; break;
        case 0x0004: tagname = "JpegThumbnailOffset"; break;
        case 0x0008: tagname = "QualityMode"; break;
        case 0x0009: tagname = "ImageSize"; break;
        case 0x000D: tagname = "FocusMode"; break;
        case 0x0014: tagname = "IsoSensitivity"; break;
        case 0x0019: tagname = "WhiteBalance"; break;
        case 0x001D: tagname = "FocalLength"; break; /* units .1mm   */
        case 0x001F: tagname = "Saturation"; break;
        case 0x0020: tagname = "Contrast"; break;
        case 0x0021: tagname = "Sharpness"; break;
        case 0x2000: tagname = "JpegThumbnail"; break;
        case 0x2011: tagname = "WhiteBalanceBias"; break;
        case 0x2012: tagname = "WhiteBalance"; break;
        case 0x2022: tagname = "ObjectDistance"; break;
        case 0x2034: tagname = "FlashDistance"; break;
        case 0x3000: tagname = "RecordMode"; break;
        case 0x3001: tagname = "SelfTimer?"; break;
        case 0x3002: tagname = "Quality"; break;
        case 0x3003: tagname = "FocusMode"; break;
        case 0x3006: tagname = "TimeZone"; break;
        case 0x3007: tagname = "BestshotMode"; break;
        case 0x3014: tagname = "CCD_ISO_Sensitivity"; break;
        case 0x3015: tagname = "ColorMode"; break;
        case 0x3016: tagname = "Enhancement"; break;
        case 0x3017: tagname = "Filter"; break;
        default: break;
    }
    return(tagname);
}

/* Report the "meaning" of tag values found at offsets in an          */
/* Asahi/Pentax MakerNote IFD entry (not at an offset).               */

void
asahi1_interpret_value(struct ifd_entry *entry_ptr)
{
    switch(entry_ptr->tag)
    {
        case 0x0001:    /* CaptureMode */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Auto"); break;
                case 1: printf("Night-scene"); break;
                case 2: printf("Manual"); break;
                case 4: printf("Multiple"); break;
                default: break;
            }
            print_endvalue();
            break;
        case 0x0002:    /* Quality */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Good"); break;
                case 1: printf("Better"); break;
                case 2: printf("Best"); break;
                default: break;
            }
            print_endvalue();
            break;
        case 0x0003:    /* Focus Mode */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 2: printf("Custom"); break;
                case 3: printf("Auto"); break;
                default: break;
            }
            print_endvalue();
            break;
        case 0x0004:    /* Flash Mode */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 1: printf("Auto"); break;
                case 2: printf("On"); break;
                case 4: printf("Off"); break;
                case 6: printf("Red-eye"); break;
                default: break;
            }
            print_endvalue();
            break;
        case 0x0007:    /* White Balance */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Auto"); break;
                case 1: printf("Daylight"); break;
                case 2: printf("Shade"); break;
                case 3: printf("Tungsten"); break;
                case 4: printf("Fluorescent"); break;
                case 5: printf("Manual"); break;
                default: break;
            }
            print_endvalue();
            break;
        case 0x000a:    /* Digital Zoom */
            if(entry_ptr->value == 0)
            {
                print_startvalue();
                printf("Off");
                print_endvalue();
            }
            break;
        case 0x000b:    /* Sharpness */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Normal"); break;
                case 1: printf("Soft"); break;
                case 2: printf("Hard"); break;
                default: break;
            }
            print_endvalue();
            break;
        case 0x000c:    /* Contrast */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Normal"); break;
                case 1: printf("Low"); break;
                case 2: printf("High"); break;
                default: break;
            }
            print_endvalue();
            break;
        case 0x000d:    /* Saturation */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Normal"); break;
                case 1: printf("Low"); break;
                case 2: printf("High"); break;
                default: break;
            }
            print_endvalue();
            break;
        case 0x0014:    /* ISO Speed Setting */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 10: printf("ISO 100"); break;
                case 16: printf("ISO 200"); break;
                case 100: printf("ISO 100"); break;
                case 200: printf("ISO 200"); break;
                default: break;
            }
            print_endvalue();
            break;
        case 0x0017:    /* Color */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 1: printf("Normal"); break;
                case 2: printf("Black & White"); break;
                case 3: printf("Sepia"); break;
                default: break;
            }
            print_endvalue();
            break;
        default:
            break;
    }
}

void
asahi2_interpret_value(struct ifd_entry *entry_ptr)
{
    char *tagname = (char *)0;

    switch(entry_ptr->tag)
    {
        case 0x0002: /* ThumbnailDimensions" */
            printf(" pixels");
            break;
        case 0x0008: /* QualityMode" */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 1: printf("Fine"); break;
                case 2: printf("Super Fine"); break;
                default: printf("unknown"); break;
            }
            print_endvalue();
            break;
        case 0x0009: /* ImageSize" */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 1: printf("640x480"); break;
                case 4: printf("1600x1200"); break;
                case 5: printf("2048x1536"); break;
                case 20: printf("2288x1712"); break;
                case 21: printf("2592x1944"); break;
                case 22: printf("2304x1728"); break;
                case 36: printf("3008x2008"); break;
                default: printf("unknown"); break;
            }
            print_endvalue();
            break;
        case 0x000D: /* FocusMode" */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 1: printf("Normal"); break;
                case 2: printf("Macro"); break;
                default: printf("unknown"); break;
            }
            print_endvalue();
            break;
        case 0x0014: /* IsoSensitivity" */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 3: printf("50"); break;
                case 4: printf("64"); break;
                case 6: printf("100"); break;
                case 9: printf("200"); break;
                default: printf("unknown"); break;
            }
            print_endvalue();
            break;
        case 0x0019: /* WhiteBalance" */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 1: printf("Fine"); break;
                case 2: printf("Shade"); break;
                case 3: printf("Tungsten"); break;
                case 4: printf("Fluorescent"); break;
                case 5: printf("Manual"); break;
                default: printf("unknown"); break;
            }
            print_endvalue();
            break;
        case 0x001D: /* FocalLength"; break; /* units .1mm   */
        case 0x001F: /* Saturation" */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("-1"); break;
                case 1: printf("Normal"); break;
                case 2: printf("+1"); break;
                default: printf("unknown"); break;
            }
            print_endvalue();
            break;
        case 0x0020: /* Contrast" */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("-1"); break;
                case 1: printf("Normal"); break;
                case 2: printf("+1"); break;
                default: printf("unknown"); break;
            }
            print_endvalue();
            break;
        case 0x0021: /* Sharpness" */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("-1"); break;
                case 1: printf("Normal"); break;
                case 2: printf("+1"); break;
                default: printf("unknown"); break;
            }
            print_endvalue();
            break;
        case 0x2012: /* WhiteBalance" */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Manual"); break;
                case 1: printf("Auto"); break;
                case 4: printf("Flash"); break;
                case 12: printf("Flash"); break;
                default: printf("unknown"); break;
            }
            print_endvalue();
            break;
        case 0x2022: /* ObjectDistance" */
            printf("mm");
            break;
        case 0x2034: /* FlashDistance" */
#if 0
Off?
            if(entry_ptr->value == 0)
            {
                print_startvalue();
                printf("Off");
                print_endvalue();
            }
#endif
            break;
        case 0x3000: /* RecordMode" */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 2: printf("Normal"); break;
                default: printf("unknown"); break;
            }
            print_endvalue();
            break;
        case 0x3001: /* SelfTimer?" */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 1: printf("Off"); break;
                default: printf("unknown"); break;
            }
            print_endvalue();
            break;
        case 0x3002: /* Quality" */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 3: printf("Fine"); break;
                default: printf("unknown"); break;
            }
            print_endvalue();
            break;
        case 0x3003: /* FocusMode" */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 1: printf("Fixed"); break;
                case 6: printf("Multi-Area Auto"); break;
                default: printf("unknown"); break;
            }
            print_endvalue();
            break;
        case 0x3007: /* BestshotMode" */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Off"); break;
                default: printf("unknown"); break;
            }
            print_endvalue();
            break;
        case 0x3015: /* ColorMode" */
#if 0
Off???
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Off"); break;
                default: printf("unknown"); break;
            }
            print_endvalue();
#endif
            break;
        case 0x3016: /* Enhancement" */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Off"); break;
                default: printf("unknown"); break;
            }
            print_endvalue();
            break;
        case 0x3017: /* Filter" */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Off"); break;
                default: printf("unknown"); break;
            }
            print_endvalue();
            break;
        default: break;
    }
}

/* Some makernotes record offsets as LONG values rather than simply   */
/* using UNDEFINED with a size; the same technique used for           */
/* the ExifIFDPointer itself, or JpegInterchangeFormat, etc.          */
/* This routine hardcodes such tags so that the print_makerentry()    */
/* routine will present them as offsets.                              */

/* Asahi has the unfortunate characteristic that tag numbers are      */
/* re-used for different purposes in the second version, and one of   */
/* those tags is an offset masquerading as a long.                    */

int
maker_asahi_is_long_offset(struct ifd_entry *entry_ptr,int model)
{
    int is_offset = 0;

    switch(model)
    {
        case ASAHI_PENTAX330RS:
        case ASAHI_PENTAX430RS:
            if(entry_ptr)
            {
                switch(entry_ptr->tag)
                {
#if 0
        /* These models also have the 0x2000 UNDEFINED thumbnail tag, */
        /* so it isn't necessary to use this kludge.                  */
                    case 0x0003: is_offset = -1; break; /* must       */
                                                        /* trigger    */
                                                        /* offset     */
                                                        /* processing */
#endif
                    case 0x0004: is_offset = 1; break;
                    default: break;
                }
            }
            break;
        case ASAHI_PENTAX330:
        case ASAHI_PENTAX430:
        default:
            break;
    }

    return(is_offset);
}
