/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/*          EXIFPROBE - TIFF/JPEG/EXIF image file probe               */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* Copyright (C) 2002 by Duane H. Hesser. All rights reserved.        */
/*                                                                    */
/* See the file LICENSE.EXIFPROBE for terms of use.                   */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#ifndef lint
static char *ModuleId = "@(#) $Id: maker_canon.c,v 1.18 2005/01/27 23:55:18 alex Exp $";
#endif

/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* Canon camera maker-specific routines                               */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* The information coded here is derived from information provided by */
/* David Burren at:                                                   */
/*     http://www.burren.cx/david/canon.html                          */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "extern.h"
#include "misc.h"
#include "maker.h"

extern void canon_camera_setting(FILE *,unsigned short,char *,
                                struct ifd_entry *,unsigned long,int);
extern void canon_exposureinfo(FILE *,unsigned short,
                                struct ifd_entry *,unsigned long,int);
extern void canon_customfunctions(FILE *,unsigned short,
                                struct ifd_entry *,unsigned long,int);


/* Find the identifying number assigned to known Canon camera models. */
/* This number is used to dispatch appropriate print and interpret    */
/* routines approopriate to the current image.                        */

int
canon_model_number(char *model,char *software)
{
    int number = GENERIC;

    if(model)
    {
        switch(*model)
        {
            case 'C':
                if(strcmp(model,"Canon DIGITAL IXUS") == 0)
                    number = CANON_IXUS;
                else if(strcmp(model,"Canon EOS D30") == 0)
                    number = CANON_EOSD30;
                else if(strcmp(model,"Canon EOS-1DS") == 0)
                    number = CANON_EOS1DS;
                else if(strcmp(model,"Canon PowerShot G2") == 0)
                    number = CANON_POWERSHOT_G2;
                break;
            default:
                number = GENERIC;
                break;
        }
    }
    return(number);
}

/* Dispatch a print routine based upon model                          */
/* Everything is (currently) mapped to one set of routines.           */

void
print_canon_makervalue(struct ifd_entry *entry_ptr,int make, int model,
                    char *prefix)
{

    if(entry_ptr && (PRINT_VALUE))
    {
        switch(model)
        {
            case CANON_IXUS:
            case CANON_EOSD30:
            case CANON_EOS1DS:
            case CANON_POWERSHOT_G2:
            case GENERIC:
            default:
                print_canon1_makervalue(entry_ptr,make,model,prefix);
                canon1_interpret_value(entry_ptr);
                break;
        }
    }
}

/* Model-specific print routine for Canon cameras. This routine is    */
/* responsible for picking off any direct entry tags which are        */
/* peculiar and will not be handled properly by print_value()         */
/* (usually UNDEFINED values which fit in the 4-byte entry value). If */
/* there are no such entries, this function simply calls              */
/* print_value().                                                     */

void
print_canon1_makervalue(struct ifd_entry *entry_ptr,int make, int model,
                    char *prefix)
{
    unsigned long value;

    if(entry_ptr && (PRINT_VALUE))
    {
        switch(entry_ptr->tag) 
        {
            case 0x000C:    /* Camera serial number                   */
                print_value(entry_ptr,prefix);
                print_startvalue();
                value = (entry_ptr->value & 0xffff0000) >> 16;
                printf("%04lX",value & 0xffff);
                value = entry_ptr->value & 0xffff;
                printf("%05lu",value & 0xffff);
                print_endvalue();
                break;
            default:
                print_value(entry_ptr,prefix);
                break;
        }
    }
}


/* Dispatch a routine to decode and print offset values for Canon     */
/* cameras.                                                           */
void
print_canon_offset_makervalue(FILE *inptr,unsigned short byteorder,
    struct ifd_entry *entry_ptr,unsigned long offset_ref,char *prefix,
    int indent,int addrwidth,int make,int model,int at_offset)
{

    if(entry_ptr && (PRINT_VALUE))
    {
        switch(model)
        {
            default:
                canon1_offset_makervalue(inptr,byteorder,entry_ptr,
                                            offset_ref,prefix,indent,
                                            addrwidth,make,model,at_offset);
                break;
        }
    }
}

/* Model-specific routine to print UNDEFINED values found at offsets  */
/* in canon makernotes.                                               */

void
canon1_offset_makervalue(FILE *inptr,unsigned short byteorder,
    struct ifd_entry *entry_ptr,unsigned long offset_ref,char *prefix,
    int indent,int addrwidth,int make,int model,int at_offset)
{
    unsigned long value_offset;
    char *nameoftag;

    if(entry_ptr)
    {
        nameoftag = maker_tagname(entry_ptr->tag,make,model); 
        value_offset = offset_ref + entry_ptr->value;
        switch(entry_ptr->tag)
        {
            case 0x0001:    /* Camera Settings                        */
                if(at_offset)
                {
                    print_tag_address(ENTRY,value_offset,indent,prefix,addrwidth);
                    print_makertagid(entry_ptr,14,": ",SUBINDENT,make,model);
                }
                printf("%lu entries",entry_ptr->count - 1);
                canon_camera_setting(inptr,byteorder,nameoftag,
                                entry_ptr,value_offset,indent + SUBINDENT);
                if(PRINT_SECTION)
                {
                    print_tag_address(SECTION,value_offset + (entry_ptr->count * 2) - 1,
                                    indent,prefix,addrwidth);
                    extraindent(SUBINDENT);
                    printf("End of %s",nameoftag);
                }
                break;
            case 0x0004:    /* Exposure Info                          */
                if(at_offset)
                {
                    print_tag_address(ENTRY,value_offset,indent,prefix,addrwidth);
                    print_makertagid(entry_ptr,12,": ",SUBINDENT,make,model);
                }
                printf("%lu entries",entry_ptr->count - 1);
                canon_exposureinfo(inptr,byteorder,
                            entry_ptr,value_offset,indent + SUBINDENT);
                if(PRINT_SECTION)
                {
                    print_tag_address(SECTION,value_offset + (entry_ptr->count * 2) - 1,
                                    indent,prefix,addrwidth);
                    extraindent(SUBINDENT);
                    printf("End of %s",nameoftag);
                }
                break;
            case 0x000F:    /* Custom Functions                       */
                if(at_offset)
                {
                    print_tag_address(ENTRY,value_offset,indent,prefix,addrwidth);
                    print_makertagid(entry_ptr,15,": ",SUBINDENT,make,model);
                }
                printf("%lu entries",entry_ptr->count - 1);
                canon_customfunctions(inptr,byteorder,
                            entry_ptr,value_offset,indent + SUBINDENT);
                if(PRINT_SECTION)
                {
                    print_tag_address(SECTION,value_offset + (entry_ptr->count * 2) - 1,
                                    indent,prefix,addrwidth);
                    extraindent(SUBINDENT);
                    printf("End of %s",nameoftag);
                }
                break;
            default:
                print_generic_offset_makervalue(inptr,byteorder,entry_ptr,
                                        offset_ref,prefix,indent,
                                        addrwidth,make,model,at_offset);
                break;
        }
    }
}


/* Canon-specific tagnames for makernotes.                            */

char *
maker_canon_tagname(unsigned short tag,int model)
{
    char *tagname = (char *)0;

    /* Check specific models first                                    */
    switch(model)
    {
        /* Add models here iff they use a different tag set           */
        default:
            break;
    }

    /* If no model-specific tag is found, use a "generic" Canon tag   */
    if(tagname == NULL)
    {
        switch(tag)
        {
            case 0x0001: tagname = "CameraSettings"; break;
            case 0x0004: tagname = "ExposureInfo"; break;
            case 0x0006: tagname = "ImageType"; break;
            case 0x0007: tagname = "FirmwareVersion"; break;
            case 0x0008: tagname = "ImageNumber"; break;
            case 0x0009: tagname = "OwnerName"; break;
            case 0x000c: tagname = "CameraSerialNo"; break;
            case 0x000f: tagname = "CustomFunctions"; break;
            default: break;
        }
    }
    return(tagname);
}

/* Report the "meaning" of tag values found directly in a Canon       */
/* MakerNote IFD entry (not at an offset).                            */

/* There are currently no known tags for which the meaning is known   */
/* so this routine is null, empty, void. There are observed tags with */
/* unknown meaning, so this routine will be necessary.                */

/* I suspect that ImageNumber could be decomposed into a DCF folder   */
/* number and sequence number),                                       */

void
canon1_interpret_value(struct ifd_entry *entry_ptr)
{
    switch(entry_ptr->tag)
    {
        default: break;
    }
}

/* Interpret the "Camera Settings" tag data from Canon MakerNotes.    */

void
canon_camera_setting(FILE *inptr,unsigned short byteorder,char *nameoftag,
    struct ifd_entry *entry_ptr,unsigned long offset,int indent)
{
    unsigned short value;
    char *sep;
    int i;

    value = read_ushort(inptr,byteorder,offset);
    offset += 2;

    if(PRINT_SECTION)
        printf(", length %u",value);
    putchar('\n');
    for(i = 1; i < entry_ptr->count; ++i)
    {
        value = read_ushort(inptr,byteorder,HERE);
        print_tag_address(ENTRY,offset,indent,"@",ADDRWIDTH);
        extraindent(SUBINDENT);
        if(!(PRINT_VALUE_AT_OFFSET))
            printf("Exif_Maker_");
        offset += 2;
        switch(i)
        {
            case 1:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Macro Mode",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"MacroMode",value);
                print_startvalue();
                switch(value)
                {
                    case 1: printf("macro"); break;
                    case 2: printf("normal"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 2:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Self Timer",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"SelfTimer",value);
                print_startvalue();
                if(value)
                    printf(".%u sec",value);
                else
                    printf("off");
                print_endvalue();
                break;
            case 3:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Quality",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"Quality",value);
                print_startvalue();
                switch(value)
                {
                    case 2: printf("normal"); break;
                    case 3: printf("fine"); break;
                    case 5: printf("super fine"); break;
                    default: printred("unknown"); break;
                }
                print_endvalue();
                break;
            case 4:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Flash Mode",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"FlashMode",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("flash did not fire"); break;
                    case 1: printf("auto"); break;
                    case 2: printf("on"); break;
                    case 3: printf("redeye"); break;
                    case 4: printf("slow sync"); break;
                    case 5: printf("auto + redeye"); break;
                    case 6: printf("on + redeye"); break;
                    case 16: printf("external flash"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 5:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Continuous Mode",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"ContinuousMode",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("single or timer"); break;
                    case 1: printf("continuous"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 6:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Unknown",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 7:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Focus Mode",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"FocusMode",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("One Shot"); break;
                    case 1: printf("AI Servo"); break;
                    case 2: printf("AI Focus"); break;
                    case 3: printf("MF"); break;
                    case 4: printf("single"); break;
                    case 5: printf("Continuous"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 8:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Unknown",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 9:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Unknown",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 10:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Image Size",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"ImageSize",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Large"); break;
                    case 1: printf("Medium"); break;
                    case 2: printf("Small"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 11:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Easy Shoot Mode",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"EasyShootMode",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Full Auto"); break;
                    case 1: printf("Manual"); break;
                    case 2: printf("Landscape"); break;
                    case 3: printf("Fast Shutter"); break;
                    case 4: printf("Slow Shutter"); break;
                    case 5: printf("Night"); break;
                    case 6: printf("B & W"); break;
                    case 7: printf("Sepia"); break;
                    case 8: printf("Portrait"); break;
                    case 9: printf("Sports"); break;
                    case 10: printf("Macro/CloseUp"); break;
                    case 11: printf("Pan Focus"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 12:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Digital Zoom",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"DigitalZoom",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("none"); break;
                    case 1: printf("2x"); break;
                    case 2: printf("4x"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 13:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Contrast",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"Contrast",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Normal"); break;
                    case 1: printf("High"); break;
                    case 0xffff: printf("Low"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 14:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Saturation",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"Saturation",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Normal"); break;
                    case 1: printf("High"); break;
                    case 0xffff: printf("Low"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 15:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Sharpness",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"Sharpness",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Normal"); break;
                    case 1: printf("High"); break;
                    case 0xffff: printf("Low"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 16:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"ISO",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"ISO",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Exif ISOSpeedRating"); break;
                    case 15: printf("Auto"); break;
                    case 16: printf("50"); break;
                    case 17: printf("100"); break;
                    case 18: printf("200"); break;
                    case 19: printf("400"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 17:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Metering Mode",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"MeteringMode",value);
                print_startvalue();
                switch(value)
                {
                    case 3: printf("Evaluative"); break;
                    case 4: printf("Partial"); break;
                    case 5: printf("Center-weighted"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 18:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Focus Type",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"FocusType",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Manual"); break;
                    case 1: printf("Auto"); break;
                    case 3: printf("Closeup (macro)"); break;
                    case 8: printf("Locked (pan mode)"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 19:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %#x",i,"AF Position Selected",value);
                else
                    printf("CS_%02d_%-18.18s = %#x",i,"AFPositionSelected",value);
                print_startvalue();
                switch(value & 0xf)
                {
                    case 0x00: printf("MF"); break;
                    case 0x01: printf("Auto"); break;
                    case 0x02: printf("Right"); break;
                    case 0x03: printf("Center"); break;
                    case 0x04: printf("Left"); break;
                    default: printred("undefined");
                }
                print_endvalue();
                break;
            case 20:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Exposure Mode",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"ExposureMode",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Easy Shooting"); break;
                    case 1: printf("Program"); break;
                    case 2: printf("Tv-priority"); break;
                    case 3: printf("Av-priority"); break;
                    case 4: printf("Manual"); break;
                    case 5: printf("A-DEP"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 21:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Unknown",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 22:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Unknown",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 23:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u units",i,"Focal Length long",value);
                else
                    printf("CS_%02d_%-18.18s = %u units",i,"FocalLength_long",value);
                break;
            case 24:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u units",i,"Focal Length short",value);
                else
                    printf("CS_%02d_%-18.18s = %u units",i,"FocalLength_short",value);
                break;
            case 25:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u per mm",i,"Focal Units",value);
                else
                    printf("CS_%02d_%-18.18s = %u per mm",i,"FocalUnits",value);
                break;
            case 26:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Unknown",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 27:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Unknown",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 28:
                if(PRINT_SECTION)
                    printf("CS_%02d_%-18.18s = %#x",i,"Flash Activity",value);
                else
                    printf("CS_%02d_%-18.18s = %#x",i,"FlashActivity",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("did not fire"); break;
                    case 1: printf("fired"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 29:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %#x ",i,"Flash Details",value);
                else
                    printf("CS_%02d_%-18.18s = %#x ",i,"FlashDetails",value);
                print_startvalue();
                sep = "";
                if(value & 0x0e)
                {
                    printf("%sexternal E-TTL",sep);
                    sep = ",";
                }
                if(value & 0x0d)
                {
                    printf("%sexternal E-TTL",sep);
                    sep = ",";
                }
                if(value & 0x0b)
                {
                    printf("%sFP sync used",sep);
                    sep = ",";
                }
                if(value & 0x07)
                {
                    printf("%s2nd-curtain sync used",sep);
                    sep = ",";
                }
                if(value & 0x04)
                {
                    printf("%sFP sync enabled",sep);
                    sep = ",";
                }
                print_endvalue();
                break;
            case 30:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Unknown",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 31:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Unknown",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 32:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %u",i,"Focus Mode",value);
                else
                    printf("CS_%02d_%-18.18s = %u",i,"FocusMode",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Single"); break;
                    case 1: printf("Continuous"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            default:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-18.18s = %#x",i,"undefined",value);
                else
                    printf("CS_%02d_%-18.18s = %#x",i,"undefined",value);
                break;
        }
        putchar('\n');
    }
}

/* Interpret the "ExposureInfo" settings in a Cannon MakerNote.       */

void
canon_exposureinfo(FILE *inptr,unsigned short byteorder,
    struct ifd_entry *entry_ptr,unsigned long offset,int indent)
{
    unsigned short value;
    int i;

    value = read_ushort(inptr,byteorder,offset);
    if(PRINT_SECTION)
        printf(", length %u",value);
    putchar('\n');
    offset += 2;
    for(i = 1; i < entry_ptr->count; ++i)
    {
        value = read_ushort(inptr,byteorder,HERE);
        print_tag_address(ENTRY,offset,indent,"@",ADDRWIDTH);
        extraindent(SUBINDENT);
        if(!(PRINT_VALUE_AT_OFFSET))
            printf("Exif_Maker_");
        offset += 2;
        switch(i)
        {
            case 1:
                if(PRINT_SECTION)
                    printf("EI_%02d:%-18.18s = %u",i,"Unknown",value);
                else
                    printf("EI_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 2:
                if(PRINT_SECTION)
                    printf("EI_%02d:%-18.18s = %u",i,"Unknown",value);
                else
                    printf("EI_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 3:
                if(PRINT_SECTION)
                    printf("EI_%02d:%-18.18s = %u",i,"Unknown",value);
                else
                    printf("EI_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 4:
                if(PRINT_SECTION)
                    printf("EI_%02d:%-18.18s = %u",i,"Unknown",value);
                else
                    printf("EI_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 5:
                if(PRINT_SECTION)
                    printf("EI_%02d:%-18.18s = %u",i,"Unknown",value);
                else
                    printf("EI_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 6:
                if(PRINT_SECTION)
                    printf("EI_%02d:%-18.18s = %u",i,"Unknown",value);
                else
                    printf("EI_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 7:
                if(PRINT_SECTION)
                    printf("EI_%02d:%-18.18s = %u",i,"White Balance",value);
                else
                    printf("EI_%02d_%-18.18s = %u",i,"WhiteBalance",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Auto"); break;
                    case 1: printf("Sunny"); break;
                    case 2: printf("Cloudy"); break;
                    case 3: printf("Tungsten"); break;
                    case 4: printf("Fluorescent"); break;
                    case 5: printf("Flash"); break;
                    case 6: printf("Custom"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 8:
                if(PRINT_SECTION)
                    printf("EI_%02d:%-18.18s = %u",i,"Unknown",value);
                else
                    printf("EI_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 9:
                if(PRINT_SECTION)
                    printf("EI_%02d:%-18.18s = %u",i,"Sequence Number",value);
                else
                    printf("EI_%02d_%-18.18s = %u",i,"SequenceNumber",value);
                break;
            case 10:
                if(PRINT_SECTION)
                    printf("EI_%02d:%-18.18s = %u",i,"Unknown",value);
                else
                    printf("EI_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 11:
                if(PRINT_SECTION)
                    printf("EI_%02d:%-18.18s = %u",i,"Unknown",value);
                else
                    printf("EI_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 12:
                if(PRINT_SECTION)
                    printf("EI_%02d:%-18.18s = %u",i,"Unknown",value);
                else
                    printf("EI_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 13:
                if(PRINT_SECTION)
                    printf("EI_%02d_%-18.18s = %u",i,"Unknown",value);
                else
                    printf("EI_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 14:
                if(PRINT_SECTION)
                    printf("EI_%02d_%-18.18s = %#x",i,"AF Position Used",value);
                else
                    printf("EI_%02d_%-18.18s = %#x",i,"AFPositionUsed",value);
                print_startvalue();
                if(value & 0xf000)
                {
                    switch(value & 0x7)
                    {
                        case 1: printf("right"); break;
                        case 2: printf("center"); break;
                        case 4: printf("left"); break;
                        default: printred("undefined"); break;
                    }
                    printf(" of %d points",(value & 0xf000) >> 12);
                }
                else
                    printf("MF");
                print_endvalue();
                break;
            case 15:
                if(PRINT_SECTION)
                    printf("EI_%02d_%-18.18s = %#x",i,"Flash Bias",value);
                else
                    printf("EI_%02d_%-18.18s = %#x",i,"FlashBias",value);
                print_startvalue();
                switch(value)
                {
                    case 0xffc0: printf("-2 EV"); break;
                    case 0xffcc: printf("-1.67 EV"); break;
                    case 0xffd0: printf("-1.50 EV"); break;
                    case 0xffd4: printf("-1.33 EV"); break;
                    case 0xffe0: printf("-1 EV"); break;
                    case 0xffec: printf("-0.67 EV"); break;
                    case 0xfff0: printf("-0.50 EV"); break;
                    case 0xfff4: printf("-0.33 EV"); break;
                    case 0x0000: printf("0 EV"); break;
                    case 0x000c: printf("+0.33 EV"); break;
                    case 0x0010: printf("+0.50 EV"); break;
                    case 0x0014: printf("+0.67 EV"); break;
                    case 0x0020: printf("+1 EV"); break;
                    case 0x002c: printf("+1.33 EV"); break;
                    case 0x0030: printf("+1.50 EV"); break;
                    case 0x0034: printf("+1.67 EV"); break;
                    case 0x0040: printf("+2 EV"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 16:
                if(PRINT_SECTION)
                    printf("EI_%02d:%-18.18s = %u",i,"Unknown",value);
                else
                    printf("EI_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 17:
                if(PRINT_SECTION)
                    printf("EI_%02d:%-18.18s = %u",i,"Unknown",value);
                else
                    printf("EI_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 18:
                if(PRINT_SECTION)
                    printf("EI_%02d:%-18.18s = %u",i,"Unknown",value);
                else
                    printf("EI_%02d_%-18.18s = %u",i,"Unknown",value);
                break;
            case 19:
                if(PRINT_SECTION)
                    printf("EI_%02d:%-18.18s = %u mm",i,"Subject Distance",value);
                else
                    printf("EI_%02d_%-18.18s = %u mm",i,"SubjectDistance",value);
                break;
            default:
                if(PRINT_SECTION)
                    printf("EI_%02d:%-18.18s = %#x",i,"Undefined",value);
                else
                    printf("EI_%02d_%-18.18s = %#x",i,"Undefined",value);
                break;
        }
        putchar('\n');
    }
}

/* Interpret Canon Custom Functions                                   */
void
canon_customfunctions(FILE *inptr,unsigned short byteorder,
    struct ifd_entry *entry_ptr,unsigned long offset,int indent)
{
    unsigned short value;
    int i,cfn,cfv;

    value = read_ushort(inptr,byteorder,offset);
    if(PRINT_SECTION)
        printf(", length %u",value);
    putchar('\n');
    offset += 2;
    for(i = 1; i < entry_ptr->count; ++i)
    {
        value = read_ushort(inptr,byteorder,HERE);
        print_tag_address(ENTRY,offset,indent,"@",ADDRWIDTH);
        extraindent(SUBINDENT);
        if(!(PRINT_VALUE_AT_OFFSET))
            printf("Exif_Maker_");
        offset += 2;
        cfn = (value & 0xff00) >> 8;
        cfv = value & 0xff;
        switch(cfn)
        {
            case 1:
                if(PRINT_SECTION)
                    printf("CF_%02d:%-30.30s = %u",cfn,"Long Exposure Noise Reduction",cfv);
                else
                    printf("CF_%02d_%-27.27s = %u",cfn,"LongExposureNoiseReduction",cfv);
                print_startvalue();
                switch(cfv)
                {
                    case 0: printf("Off"); break;
                    case 1: printf("On"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 2:
                if(PRINT_SECTION)
                    printf("CF_%02d:%-30.30s = %u",cfn,"Shutter AE lock Buttons",cfv);
                else
                    printf("CF_%02d_%-27.27s = %u",cfn,"ShutterAElockButtons",cfv);
                print_startvalue();
                switch(cfv)
                {
                    case 0: printf("AF/AE lock"); break;
                    case 1: printf("AE lock/AF"); break;
                    case 2: printf("AF/AF lock"); break;
                    case 3: printf("AE+release/AE+AF"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 3:
                if(PRINT_SECTION)
                    printf("CF_%02d:%-30.30s = %u",cfn,"Mirror Lockup",cfv);
                else
                    printf("CF_%02d_%-27.27s = %u",cfn,"MirrorLockup",cfv);
                print_startvalue();
                switch(cfv)
                {
                    case 0: printf("Disabled"); break;
                    case 1: printf("Enabled"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 4:
                if(PRINT_SECTION)
                    printf("CF_%02d:%-30.30s = %u",cfn,"Tv/Av And Exposure Level",cfv);
                else
                    printf("CF_%02d_%-27.27s = %u",cfn,"TvAvAndExposureLevel",cfv);
                print_startvalue();
                switch(cfv)
                {
                    case 0: printf("1/2 stop"); break;
                    case 1: printf("1/3 stop"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 5:
                if(PRINT_SECTION)
                    printf("CF_%02d:%-30.30s = %u",cfn,"AF Assist Light",cfv);
                else
                    printf("CF_%02d_%-27.27s = %u",cfn,"AFAssistLight",cfv);
                print_startvalue();
                switch(cfv)
                {
                    case 0: printf("On (auto)"); break;
                    case 1: printf("Off"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 6:
                if(PRINT_SECTION)
                    printf("CF_%02d:%-30.30s = %u",cfn,"Shutter Speed In AV Mode",cfv);
                else
                    printf("CF_%02d_%-27.27s = %u",cfn,"ShutterSpeedInAVMode",cfv);
                print_startvalue();
                switch(cfv)
                {
                    case 0: printf("Auto"); break;
                    case 1: printf("1/200 (fixed)"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 7:
                if(PRINT_SECTION)
                    printf("CF_%02d:%-30.30s = %u",cfn,"AEB Sequence Auto Cancellation",cfv);
                else
                    printf("CF_%02d_%-27.27s = %u",cfn,"AEBSequenceAutoCancellation",cfv);
                print_startvalue();
                switch(cfv)
                {
                    case 0: printf("0,-,+/Enabled"); break;
                    case 1: printf("0,-,+/Disabled"); break;
                    case 2: printf("-,0,+/Enabled"); break;
                    case 3: printf("-,0,+/Disabled"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 8:
                if(PRINT_SECTION)
                    printf("CF_%02d:%-30.30s = %u",cfn,"Shutter Curtain Sync",cfv);
                else
                    printf("CF_%02d_%-27.27s = %u",cfn,"ShutterCurtainSync",cfv);
                print_startvalue();
                switch(cfv)
                {
                    case 0: printf("1st-curtain sync"); break;
                    case 1: printf("2nd-curtain sync"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 9:
                if(PRINT_SECTION)
                    printf("CF_%02d:%-30.30s = %u",cfn,"Lens AF Stop Button Fn Switch",cfv);
                else
                    printf("CF_%02d_%-27.27s = %u",cfn,"LensAFStopButtonFnSwitch",cfv);
                print_startvalue();
                switch(cfv)
                {
                    case 0: printf("AF Stop"); break;
                    case 1: printf("Operate AF"); break;
                    case 2: printf("Lock AE and start timer"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 10:
                if(PRINT_SECTION)
                    printf("CF_%02d:%-30.30s = %u",cfn,"Auto Reduction Of Fill Flash",cfv);
                else
                    printf("CF_%02d_%-27.27s = %u",cfn,"AutoReductionOfFillFlash",cfv);
                print_startvalue();
                switch(cfv)
                {
                    case 0: printf("Enabled"); break;
                    case 1: printf("Disabled"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 11:
                if(PRINT_SECTION)
                    printf("CF_%02d_%-30.30s = %u",cfn,"Menu Button Return Position",cfv);
                else
                    printf("CF_%02d_%-27.27s = %u",cfn,"MenuButtonReturnPosition",cfv);
                print_startvalue();
                switch(cfv)
                {
                    case 0: printf("Top"); break;
                    case 1: printf("Previous (volatile)"); break;
                    case 2: printf("Previous"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 12:
                if(PRINT_SECTION)
                    printf("CF_%02d:%-30.30s = %u",cfn,"SET Button Func When Shooting",cfv);
                else
                    printf("CF_%02d_%-27.27s = %u",cfn,"SETButtonFuncWhenShooting",cfv);
                print_startvalue();
                switch(cfv)
                {
                    case 0: printf("Not assigned"); break;
                    case 1: printf("Change quality"); break;
                    case 2: printf("Change ISO speed"); break;
                    case 3: printf("PreviouSelect parameters"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 13:
                if(PRINT_SECTION)
                    printf("CF_%02d:%-30.30s = %u",cfn,"Sensor Cleaning",cfv);
                else
                    printf("CF_%02d_%-27.27s = %u",cfn,"SensorCleaning",cfv);
                print_startvalue();
                switch(cfv)
                {
                    case 0: printf("Disabled"); break;
                    case 1: printf("Enabled"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            default:
                if(PRINT_SECTION)
                    printf("CF_%02d:%-30.30s = %#x",cfn,"undefined",cfv);
                else
                    printf("CF_%02d_%-27.27s = %#x",cfn,"undefined",cfv);
                break;
        }
        putchar('\n');
    }
}
