/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/*          EXIFPROBE - TIFF/JPEG/EXIF image file probe               */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* Copyright (C) 2002 by Duane H. Hesser. All rights reserved.        */
/*                                                                    */
/* See the file LICENSE.EXIFPROBE for terms of use.                   */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#ifndef lint
static char *ModuleId = "@(#) $Id: maker_generic.c,v 1.29 2005/01/28 18:02:38 alex Exp $";
#endif

/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* "Generic" routines for makernotes which use TIFF IFD structure     */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "extern.h"
#include "misc.h"
#include "maker.h"

/* Read a makernote at the offset given in the passed maker_entry.    */

/* Checks for data with a TIFF IFD structure at the makernote offset, */
/* possibly offset by a maker name string and/or a new TIFF header    */
/* (e.g. some later Nikons). Notes which pass the test are scanned    */
/* and printed. If the note is in IFD format but the maker is         */
/* unknown, a "bare bones" entry will be printed, with tag numbers,   */
/* type, etc. If the maker is known, tag names, undefined values and  */
/* "interpretations" of the data may be printed. MakerNotes with      */
/* unknown/non-ifd structures could be dispatched to specialized      */
/* routines, but at this time I do not have information on any of     */
/* those. Unknown notes (or notes with invalid IFD entries) can be    */
/* dumped in hex/ascii by user option (-M). 'Undefined" data may be   */
/* similarly dumped by the -U option, if maker-specific routines      */
/* don't handle it.                                                   */

/* The return is the offset of the end of the note; if the note       */
/* cannot be read, return is 0L.                                      */

unsigned long
process_makernote(FILE *inptr,unsigned short byteorder,
                    struct ifd_entry *maker_entry,
                    unsigned long fileoffset_base,
                    unsigned long max_offset,
                    int indent,int addrwidth)
{
    struct ifd_entry *entry_ptr;
    struct maker_scheme *maker_scheme;
    unsigned long maker_offset = 0UL;
    unsigned long maker_count = 0UL;
    unsigned long max_value_offset = 0UL;
    unsigned long start_offset = 0UL;
    unsigned long dumplength = 0UL;
    unsigned long end_of_note = 0UL;
    unsigned long next_ifd_offset = 0UL;
    unsigned long current_offset;
    unsigned long start_entry_offset;
    unsigned long entry_offset,value_offset;
    char *prefix,*swname;
    int entry_num,num_entries,make,model;
    int invalid = 0;
    int value_is_offset = 0;
    int note_relative_offset_scheme = 0;
    int use_second_pass = 0;

    if(inptr && maker_entry)
    {
        STARTCOLOR(MAKER_COLOR);
        /* Save these; they're in static storage in the maker_entry   */
        /* and will be overwritten when entry reading starts.         */
        maker_offset = maker_entry->value;
        maker_count = maker_entry->count;
        if(maker_count > 4UL)
            end_of_note = maker_offset + maker_count + fileoffset_base;
        else if(maker_count == 4UL)
        {
            /* special case, for certain Casio MakerNotes which use   */
            /* 4-byte UNDEFINED as though it were a LONG offset       */
            if(maker_entry->value_type == UNDEFINED)
                maker_offset = to_ulong((unsigned char *)&maker_entry->value,byteorder);
            /* This pretty much forces maker_scheme to PLAIN          */
            maker_count=0L;
        }
        else
            goto cant_dumpit;

        /* Get make and model from the global names saved from IFD0   */
        /* Override real names if -m, -l, -s                          */
        if(Use_Make_name)
            make = maker_number(Use_Make_name);
        else
            make = maker_number(Make_name);

        if(Use_Software_name)
            swname = Use_Software_name;
        else
            swname = Software_name;

        if(Use_Model_name)
            model = model_number(make,Use_Model_name,swname);
        else
            model = model_number(make,Model_name,swname);


        /* start_offset is where the note begins, relative to the     */
        /* start of the file.                                         */
        start_offset = start_entry_offset = maker_offset + fileoffset_base;

        /* Mark the beginning offset of the note info. At least one   */
        /* maker writes the makernote past the end of parent IFD so   */
        /* that it is not actually contained within the parent (EXIF) */
        /* IFD. The following attempts to mark that situation by      */
        /* using a '^' to mark the start address, so that it stands   */
        /* out a little. If "max_offset" is non-zero, it is presumed  */
        /* to be the highest offset of the parent, ususally taken     */
        /* from a 'next_ifd_offset' field.                            */
        if(max_offset && (start_offset > max_offset))
            prefix = "^";
        else
            prefix = "@";
        if((PRINT_LIST))
        {
            /* In list mode, when SECTIONS are not printed, provide   */
            /* these items as "pseudotags".                           */
            print_tag_address(ENTRY,start_offset,indent + SUBINDENT,prefix,addrwidth);
            /* those old Casio notes...                               */
            if(maker_count == 0UL)
                printf("%-*.*s = %lu\n",TAGWIDTH + PREFIXLEN,
                        TAGWIDTH + PREFIXLEN,"Exif_MakerNoteOffset",
                        start_offset);
            printf("%-*.*s = %lu\n",TAGWIDTH + PREFIXLEN,
                    TAGWIDTH + PREFIXLEN,"Exif_MakerNoteCount",
                    maker_count ? maker_count: maker_entry->count);
            print_tag_address(ENTRY,start_offset,indent + SUBINDENT,prefix,addrwidth);
            printf("%-*.*s = ",TAGWIDTH + PREFIXLEN,
                    TAGWIDTH + PREFIXLEN,"Exif_MakerNoteScheme");
        }
        else if(PRINT_SECTION)
        {
            print_tag_address(SECTION,start_offset,indent + SUBINDENT,prefix,addrwidth);
            printf("<MakerNote> count %lu, ",
                    maker_count ? maker_count: maker_entry->count);
        }

        /* Discover, if possible, the 'scheme' by which the MakerNote */
        /* is stored.                                                 */
        maker_scheme = find_makerscheme(inptr,maker_offset,fileoffset_base,
                                            byteorder,maker_count,make,model);

        /* If scheme debugging is in effect, the type info below      */
        /* needs to be indented a bit, since the debug output splits  */
        /* the starting <MakerNote> line.                             */
        if(PRINT_SCHEME_DEBUG)
            putindent(indent + SUBINDENT + SUBINDENT + SUBINDENT);
        /* Print a brief description of the scheme found. Enclose it  */
        /* in quotes for LIST mode.                                   */
        if((PRINT_LIST))
            putchar('"');
        switch(maker_scheme->scheme_type)
        {
            case HAS_ID_SCHEME:
                if((PRINT_LIST) || (PRINT_SECTION))
                {
                    if(maker_scheme->offsets_relative_to == START_OF_NOTE)
                        printf("Note-relative offset ");
                    else if(maker_scheme->offsets_relative_to == START_OF_ENTRY)
                        printf("Entry-relative offset ");
                    printf("ID");
                }
                break;
            case HAS_ID_PLUS_TIFFHDR_SCHEME:
                if((PRINT_LIST) || (PRINT_SECTION))
                    printf("ID plus TIFFheader");
                break;
            case PLAIN_IFD_SCHEME:
                if((PRINT_LIST) || (PRINT_SECTION))
                {
                    if(maker_scheme->offsets_relative_to == START_OF_NOTE)
                        printf("Note-relative offset ");
                    else if(maker_scheme->offsets_relative_to == START_OF_ENTRY)
                        printf("Entry-relative offset ");
                    else
                        printf("Normal ");
                    printf("IFD");
                    if(maker_count == 0) /* some old Casio                */
                        printf(" from LONG offset");
                }
                break;
            /* New schemes can be added here.                         */
            case UNKNOWN_SCHEME:
            default:
                /* dump a couple of lines, unless user says otherwise */
                if((PRINT_LIST) || (PRINT_SECTION))
                {
                    next_ifd_offset = 0UL;
                    printf("unknown");
                    if(PRINT_SECTION)
                        printf(" scheme");
                    if(PRINT_SECTION && !(PRINT_MAKERDUMP))
                    {
                        (void)putchar('\n');
                        hexdump(inptr,start_offset,maker_count,24,12,
                                    indent,SUBINDENT,addrwidth);
                        (void)putchar('\n');
                    }
                    else
                        printf("\"\n");
                }
                if(PRINT_MAKERNOTE_SCHEME)
                    print_scheme(inptr,maker_scheme);
                goto dumpit;
                break;
        }
        if((PRINT_LIST))
            printf("\"\n");
        else if(PRINT_SECTION)
            printf(" scheme,"); /* there's more to print on this line */
        num_entries = maker_scheme->num_entries;
        byteorder = maker_scheme->note_byteorder;
        if((maker_scheme->scheme_type == HAS_ID_PLUS_TIFFHDR_SCHEME) ||
                (maker_scheme->offsets_relative_to == START_OF_NOTE))
        {
            fileoffset_base = maker_scheme->fileoffset_base;
            start_entry_offset = fileoffset_base + maker_scheme->ifd_offset + 2;
        }
        else
            start_entry_offset = start_offset + maker_scheme->ifd_offset + 2;
        if(PRINT_SECTION)
            printf(" %d entries starting at offset %#lx/%lu\n",num_entries,
                        start_entry_offset,start_entry_offset);
        if(PRINT_MAKERNOTE_SCHEME)
            print_scheme(inptr,maker_scheme);

#ifdef SHOW_MAKERID
        extraindent(SUBINDENT);
        if(PRINT_SECTION)
            printf("{ Make_name=%s (ep maker #%d); Model_name=%s, Software_name=%s (ep model #%d) }\n",
                        Make_name?Make_name:"",make,Model_name?Model_name:"",
                        Software_name?Software_name:"",model);
#endif

        if((maker_scheme->id_length > 0) && (PRINT_SECTION))
        {
            print_tag_address(SECTION,start_offset, indent+SUBINDENT,
                                    prefix,addrwidth);
            extraindent(SUBINDENT);
            printf("MakerId=\"%s\"\n",maker_scheme->id);
        }

        if(((maker_scheme->scheme_type == HAS_ID_PLUS_TIFFHDR_SCHEME)) &&
                    (PRINT_SECTION))
        {
            print_tag_address(SECTION,fileoffset_base,indent+SUBINDENT,
                                    prefix,addrwidth);
            extraindent(SUBINDENT);
            if(print_header(maker_scheme->note_byteorder,SECTION) == 0)
                printf(" ifd offset = %#lx/%ld)\n",maker_scheme->ifd_offset,
                        maker_scheme->ifd_offset);
        }

        /* At this point, the note has been "verified" to be in TIFF  */
        /* IFD format, and can be read normally in that fashion.      */
        /* Maker-specific information is handled in the print_maker   */
        /* routines, according to the make and model numbers.         */

        /* Scheme validation is not entirely robust, however, so the  */
        /* code below bails out at the first sign of an invalid       */
        /* entry, and hexdumps the data if wanted.                    */

        current_offset = entry_offset = start_entry_offset;
        value_is_offset = 0;
        for(entry_num = 0; entry_num < num_entries; ++entry_num)
        {
            print_tag_address(ENTRY,entry_offset,indent+SUBINDENT,prefix,addrwidth);
            entry_ptr = read_entry(inptr,byteorder,entry_offset);
            if((entry_ptr->value_type == 0) || (entry_ptr->value_type > DOUBLE) ||
                ferror(inptr) || feof(inptr))
            if(invalid)
            {
                extraindent((addrwidth * 2) + 6 + indent + (SUBINDENT * 2));
                if(PRINT_SECTION)
                    printf("INVALID MAKERNOTE ENTRY %d at offset %lu\n",
                            invalid,entry_offset);
                goto dumpit;
            }
            if(ferror(inptr))
                goto dumpit;
            if(maker_scheme->offsets_relative_to == START_OF_ENTRY)
                fileoffset_base = entry_offset;
            entry_offset = current_offset = ftell(inptr);
            value_offset = print_makerentry(inptr,byteorder,addrwidth,
                                            entry_ptr,fileoffset_base,
                                                indent,make,model);
            if(value_offset == 0UL)
                value_offset = entry_offset;
            if(value_offset > max_value_offset)
                max_value_offset = value_offset;
            if(is_offset(entry_ptr) && (PRINT_VALUE_AT_OFFSET))
                ++use_second_pass;
        }

        /* Some makers write the next ifd offset as 0, as is normal   */
        /* for an IFD. Some write nothing.                            */
        if(maker_scheme->no_nextifd_offset)
        {
            next_ifd_offset = 0L;
            if((PRINT_ADDRESS) && (PRINT_VALUE) && (PRINT_VALUE_AT_OFFSET))
            {
                /* Since it isn't needed, just note it and move on.       */
                print_tag_address(ENTRY,current_offset,indent+SUBINDENT,prefix,addrwidth);
                extraindent(SUBINDENT);
                printred("no next IFD offset\n");
            }
        }
        else
        {
            next_ifd_offset = read_ulong(inptr,byteorder,current_offset);
            if(next_ifd_offset > 0L)
            {
                /* This should not happen, of course.                     */
                if((PRINT_ADDRESS) && (PRINT_VALUE) && (PRINT_VALUE_AT_OFFSET))
                {
                    print_tag_address(ENTRY,current_offset,indent+SUBINDENT,prefix,addrwidth);
                    extraindent(SUBINDENT);
                    printf("**** next IFD offset %lu  ",next_ifd_offset);
                    next_ifd_offset += fileoffset_base;
                    printf("(+ %lu = %#lx/%lu)\n",fileoffset_base,
                                next_ifd_offset,next_ifd_offset);
                    if(next_ifd_offset < ftell(inptr))
                    {
                        printred("BAD NEXT IFD OFFSET\n");
                        next_ifd_offset = 0L;
                    }
                    else
                        putchar('\n');
                }
                else
                    next_ifd_offset += fileoffset_base;
            }
            else
            {
                if((PRINT_ADDRESS) && (PRINT_VALUE) && (PRINT_VALUE_AT_OFFSET))
                {
                    print_tag_address(ENTRY,current_offset,indent+SUBINDENT,prefix,addrwidth);
                    extraindent(SUBINDENT);
                    printf("**** next IFD offset 0\n");
                }
            }
        }

        if(ferror(inptr))
        {
            printf(" READ NEXT IFD OFFSET FAILED ");
            why(stdout);
            clearerr(inptr);    /* keep going...                      */
        }
        else
            current_offset = ftell(inptr);

        value_offset = current_offset;

        /* Make a second pass over the entries, to print values of    */
        /* entries whose values are indirect.                         */
        if(use_second_pass)
        {

            if((PRINT_VALUE) && (PRINT_ADDRESS) && (PRINT_VALUE_AT_OFFSET))
            {
                print_tag_address(ENTRY,value_offset,indent+SUBINDENT,prefix,addrwidth);
                extraindent(SUBINDENT);
                printf("============= VALUES, MakerNote ============\n");
            }

            entry_offset = start_entry_offset;
            for(entry_num = 0; entry_num < num_entries; ++entry_num)
            {
                entry_ptr = read_entry(inptr,byteorder,entry_offset);
                if(ferror(inptr))   /* why didn't this happen above?  */
                    goto dumpit;
                if(maker_scheme->offsets_relative_to == START_OF_ENTRY)
                    fileoffset_base = entry_offset;
                entry_offset = current_offset = ftell(inptr);
                if(is_offset(entry_ptr) || maker_is_long_offset(entry_ptr,make,model))
                {
                    if((PRINT_VALUE_AT_OFFSET))
                        value_offset =
                            print_offset_makervalue(inptr,byteorder,
                                            entry_ptr,fileoffset_base,
                                            prefix,indent+SUBINDENT,
                                            addrwidth,make,model,1);
                    else
                        value_offset = fileoffset_base + entry_ptr->value + 
                            (value_type_size(entry_ptr->value_type) * entry_ptr->count);
                    if(value_offset > max_value_offset) 
                        max_value_offset = value_offset;

                    current_offset = ftell(inptr);
                    if(current_offset > max_value_offset)
                        max_value_offset = current_offset;
                }
            }
        }
        else if(current_offset > max_value_offset)
            max_value_offset = current_offset;

        /* If the note is incompletely specified, point that out. If  */
        /* IFD entries are not being printed this message is not      */
        /* wanted (and max_value_offset is likely 0 anyway).          */
        /* This actually happens...                                   */
        if((end_of_note && (max_value_offset < end_of_note)) && (PRINT_ENTRY) &&
            !(PRINT_LIST))
        {
            unsigned long printlength = 0;

            print_tag_address(ENTRY,max_value_offset - 1,indent + SUBINDENT,"*",addrwidth);
            extraindent(SUBINDENT);
            printf("---- End of values before end of Makernote\n");
            dumplength = end_of_note - max_value_offset;
            /* dump at least the first couple of rows, just to see... */
            if(PRINT_MAKERDUMP)
            {
                if((Max_makerdump == DUMPALL) || (Max_makerdump > dumplength))
                    printlength = dumplength;
                else if(Max_makerdump > 0L)
                    printlength = Max_makerdump;
                else if(Max_undefined > 0L)
                    printlength = Max_undefined;
            }
            else    /* always print a little bit in this case         */
                printlength = dumplength > 48 ? 48 : dumplength;
            hexdump(inptr,max_value_offset,dumplength,printlength,12,
                            indent,SUBINDENT,addrwidth);
            (void)putchar('\n');
        }
        else if(maker_count == 0L)
            end_of_note = max_value_offset;
            
        print_tag_address(SECTION,end_of_note - 1,indent + SUBINDENT,"-",addrwidth);
        if(max_offset && (start_offset > max_offset))
            end_of_note = 0L;
        if(PRINT_SECTION)
            printf("</MakerNote>");
        if(next_ifd_offset)
            printf("# WARNING! MakerNote has next ifd offset %lu\n",next_ifd_offset);
        else if(PRINT_SECTION)
            (void)putchar('\n');
        STOPCOLOR;
    }
    return(end_of_note);

dumpit:
    if(PRINT_MAKERDUMP)
    {
        /* If no specific length is given for makernote data, limit   */
        /* the dump length to the length specified for undefined      */
        /* data.                                                      */
        if((Max_makerdump == DUMPALL) || (Max_makerdump > maker_count))
            dumplength = maker_count;
        else if(Max_makerdump > 0L)
            dumplength = Max_makerdump;
        else
            dumplength = Max_undefined;
        if(dumplength > 0L)
        {
            (void)putchar('\n');
#ifdef SHOW_MAKERID
            extraindent(SUBINDENT);
            printf("{ Make_name=%s (ep maker #%d); Model_name=%s, Software_name=%s (ep model #%d) }\n",
                        Make_name?Make_name:"",make,Model_name?Model_name:"",
                        Software_name?Software_name:"",model);
#endif
            hexdump(inptr,start_offset,maker_count,dumplength,12,
                    indent,SUBINDENT,addrwidth);
            (void)putchar('\n');
        }
    }
    else
    {
        if((PRINT_SECTION) && (maker_count > 24))
        {
            putindent(2 * (addrwidth + indent + SUBINDENT + SUBINDENT + SUBINDENT));
            printred(" (use -M to see more)\n");
        }
#ifdef SHOW_MAKERID
            extraindent(SUBINDENT);
        if(PRINT_SECTION)
            printf("{ Make_name=%s (ep maker #%d); Model_name=%s, Software_name=%s (ep model #%d) }\n",
                        Make_name?Make_name:"",make,Model_name?Model_name:"",
                        Software_name?Software_name:"",model);
#endif
    }
    if(PRINT_SECTION)
    {
        print_tag_address(SECTION,end_of_note - 1,indent + SUBINDENT,"-",addrwidth);
        printf("</MakerNote>\n");
    }
    STOPCOLOR;
    if(maker_count)
        return(end_of_note);
    return(0L);
cant_dumpit:
    printf(" INVALID MakerNote offset %lu\n",maker_offset);
    return(0L);
}

/* Return a unique number for a given maker, (assigned in "maker.h")  */
/* based upon a string, usually taken from a "Make" tag. This "magic  */
/* number" will be used to select routines to print and interpret     */
/* data from MakerNotes for that maker.                               */

int
maker_number(char *make)
{
    int number = GENERIC;

    if(make)
    {
        /* Fast first match, then string compare                      */
        switch(*make)
        {
            case 'a':
            case 'A':
                if(strncasecmp(make,"agfa",4) == 0)
                    number = MAKER_AGFA;
                else if(strncasecmp(make,"asahi",5) == 0)
                    number = MAKER_ASAHI;   /* Pentax                 */
                /* else add makers here                               */
                break;
            case 'c':
            case 'C':
                if(strncasecmp(make,"canon",5) == 0)
                    number = MAKER_CANON;
                else if(strncasecmp(make,"casio",5) == 0)
                    number = MAKER_CASIO;
                /* else add makers here                               */
                break;
            case 'e':
            case 'E':
                if(strncasecmp(make,"eastman",7) == 0)
                    number = MAKER_KODAK;
                /* else add makers here                               */
                break;
            case 'f':
            case 'F':
                if(strncasecmp(make,"fuji",4) == 0)
                    number = MAKER_FUJIFILM;
                /* else add makers here                               */
                break;
            case 'h':
            case 'H':
                if(strncasecmp(make,"hewlett",7) == 0)
                    number = MAKER_HP;
                /* else add makers here                               */
                break;
            case 'k':
            case 'K':
                if(strncasecmp(make,"konica",6) == 0)
                    number = MAKER_KONICA;
                else if(strncasecmp(make,"kyocera",7) == 0)
                    number = MAKER_KYOCERA;
                /* else add makers here                               */
                break;
            case 'l':
            case 'L':
                if(strncasecmp(make,"leica",5) == 0)
                    number = MAKER_LEICA;
                /* else add makers here                               */
                break;
            case 'm':
            case 'M':
                if(strncasecmp(make,"minolta",7) == 0)
                    number = MAKER_MINOLTA;
                /* else add makers here                               */
                break;
            case 'p':
            case 'P':
                if(strncasecmp(make,"panasonic",9) == 0)
                    number = MAKER_PANASONIC;
                /* else add makers here                               */
                break;
            case 'n':
            case 'N':
                if(strncasecmp(make,"nikon",5) == 0)
                    number = MAKER_NIKON;
                /* else add makers here                               */
                break;
            case 'o':
            case 'O':
                if(strncasecmp(make,"olympus",7) == 0)
                    number = MAKER_OLYMPUS;
                /* else add makers here                               */
                break;
            case 'r':
            case 'R':
                if(strncasecmp(make,"ricoh",5) == 0)
                    number = MAKER_RICOH;
                /* else add makers here                               */
                break;
            case 's':
            case 'S':
                if(strncasecmp(make,"sony",4) == 0)
                    number = MAKER_SONY;
                else if(strncasecmp(make,"sanyo",5) == 0)
                    number = MAKER_SANYO;
                else if(strncasecmp(make,"seiko",5) == 0)
                    number = MAKER_EPSON;
                else if(strncasecmp(make,"sigma",5) == 0)
                    number = MAKER_SIGMA;
                /* else add makers here                               */
                break;
            case 't':
            case 'T':
                if(strncasecmp(make,"toshiba",8) == 0)
                    number = MAKER_TOSHIBA;
                if(strncasecmp(make,"traveler",8) == 0)
                    number = MAKER_TRAVELER;
                /* else add makers here                                   */
                break;
            default:
                break;
        }
    }
    return(number);
}

/* Call a maker-specific routine to assign an identifying number for  */
/* a specific model of a given maker's camera. The camera make is     */
/* identified by the number assigned by make_number(), and is used to */
/* dispatch the maker-specific routine, which is then free to pick an */
/* identifying number based upon the "Model" tag and possibly the     */
/* "Software" tag. The numbers are assigned in "maker.h" and will be  */
/* used to select print routines appropriate to the makernote data    */
/* for the current image. It is possible to assign a single number to */
/* a group of models, or to assign a separate number to each          */
/* model/software combination, requiring the maker-specific routines  */
/* to dispatch the appropriate routines.                              */

int
model_number(int make,char *model,char *software)
{
    int number = GENERIC;

    if(model == NULL)
        model = NULLSTRING;
    if(software == NULL)
        software = NULLSTRING;
    switch(make)
    {
        case MAKER_AGFA:
            number = agfa_model_number(model,software);
            break;
        case MAKER_ASAHI:   /* PENTAX                                 */
            number = asahi_model_number(model,software);
            break;
        case MAKER_CANON:
            number = canon_model_number(model,software);
            break;
        case MAKER_CASIO:
            number = casio_model_number(model,software);
            break;
        case MAKER_EPSON:
            number = epson_model_number(model,software);
            break;
        case MAKER_FUJIFILM:
            number = fujifilm_model_number(model,software);
            break;
        case MAKER_HP:      /* no makernote data                      */
            break;
        case MAKER_KODAK:   /* no makernote data                      */
            break;
        case MAKER_KONICA:  /* no makernote data                      */
            break;
        case MAKER_KYOCERA:
            number = kyocera_model_number(model,software);
            break;
        case MAKER_LEICA:
            number = leica_model_number(model,software);
            break;
        case MAKER_MINOLTA:
            number = minolta_model_number(model,software);
            break;
        case MAKER_NIKON:
            number = nikon_model_number(model,software);
            break;
        case MAKER_OLYMPUS:
            number = olympus_model_number(model,software);
            break;
        case MAKER_PANASONIC:
            number = panasonic_model_number(model,software);
            break;
        case MAKER_RICOH:    /* no or non-IFD makernote data          */
            break;
        case MAKER_SANYO:
            number = sanyo_model_number(model,software);
            break;
        case MAKER_SIGMA:
            number = sigma_model_number(model,software);
            break;
        case MAKER_SONY:    /* no makernote data                      */
            /* DSC-V3 has makernote in IFD format, but nothing is     */
            /* known about meaning (there are not maker-specific      */
            /* routines here yet)                                     */
            break;
        case MAKER_TOSHIBA:
            number = toshiba_model_number(model,software);
            break;
        case MAKER_TRAVELER:    /* proprietary makernote              */
            break;
        default:    /* GENERIC                                        */
            break;
    }
    return(number);
}

/* *********************************************************************/
/* Attempt to discover the format of a MakerNote. Some notes are in   */
/* normal TIFF IFD format, with tags assigned by the maker, and       */
/* values interpreted in maker-specific ways. Some makers begin with  */
/* a few bytes containing an ascii string identifying the maker, with */
/* a normal IFD following at a specific offset. One camera I've seen  */
/* begins with such an ascii id, followed by an actual TIFF header,   */
/* with the following IFD offsets then relative to that TIFF header.  */
/* This latter scheme seems the most sensible, since it allows image  */
/* editting software to move the MakerNote with little effort, and    */
/* allows standard TIFF machinery to read them.                       */

/* Within the three schemes above, all of which include the basic     */
/* format of a TIFF Image File Directory, makers have devised several */
/* ways to express their individuality.                               */

/* The most harmless is the length of the ID string, and the          */
/* resulting offset from the start of the MakerNote blob to the start */
/* of the actual IFD. Some makers prove their independence by writing */
/* "offset" entries in the IFD (required when a value won't fit in    */
/* the 12-byte entry) relative to the start of the MakerNote or even  */
/* the start of the current entry. The TIFF specification requires    */
/* that such offsets are relative to the beginning of the last        */
/* in-scope TIFF header.                                              */

/* At least one maker thumbs it's nose at the software world by       */
/* gratuitously changing the byte order used when writing the         */
/* MakerNote, compared to the remainder of the image file.            */

/* Some makers write MakerNotes, but in a purely private, proprietary */
/* format.  We cannot read those, except as a blob of useless junk.   */

/* This routine uses some simple heuristics to attempt to intuit the  */
/* format of a note, rather than a table of known make/model formats. */
/* If an Id string is found, an attempt is made to find a TIFF header */
/* or an apparently valid IFD. If found, the offset from the start of */
/* the blob is recorded, and if the byteorder is changed, that is     */
/* noted as well                                                      */

/* Offset to the start of the IFD and byteorder are sufficient to     */
/* read the directory entries (and print them) but the reference      */
/* point for offsets to indirect data is difficult to determine if it */
/* differs from the TIFF specification. Specific information,         */
/* according to camera make and model, is added for cameras which are */
/* known to write note- or entry-relative offsets. Some cameras also  */
/* fail to write zeros for the "next IFD offset" normally included at */
/* the end of a TIFF IFD, so that fact is noted as well.              */

/* If the note just doesn't look presentable, a (currently dummy)     */
/* routine can be used to check for known sports.                     */

/* The information recorded in the maker_scheme structure at this     */
/* point is sufficient to read and print the structure of the note    */
/* (except for UNKNOWN schemes, of course). Interpreting the          */
/* *meaning* of the tags and values is the responsibility of          */
/* maker-specific routines.                                           */
/* *********************************************************************/

/* *********************************************************************/
/* The maker_scheme structure may be of interest to mnaker-specific   */
/* routines, but is not passed around, and should not be modified     */
/* ooutside find_makerscheme. retrieve_makerscheme is provided to     */
/* allow a *copy* of the structure to be retrieved after it has been  */
/* filled.                                                            */
/* *********************************************************************/

/* Allow only find_makerscheme() to muck with the original scheme     */
static struct maker_scheme maker_scheme;

struct maker_scheme *
find_makerscheme(FILE *inptr,unsigned long note_offset,
                    unsigned long fileoffset_base,
                    unsigned short byteorder,
                    unsigned long maker_size,int make,int model)
{
    int length_read,id_length,num_entries,start,maxchk_offset;
    int tiffhdr_offset = 0;
    unsigned long note_fileoffset = 0L;
    unsigned long num_entry_offset = 0L;
    unsigned long ifd_offset = 0L;
    unsigned long scheme_type = UNKNOWN_SCHEME;
    unsigned short marker,alt_byteorder,note_byteorder;
    char *id;

    memset(&maker_scheme,'\0',sizeof(struct maker_scheme));
    note_fileoffset = note_offset + fileoffset_base;

    id= (char *)read_bytes(inptr,MAX_IDLEN,note_fileoffset);
    length_read = strlen(id);

    /* check that string is printable, adjust id_length if nexessary  */
    if(PRINT_SCHEME_DEBUG)
    {
        pushcolor(HI_RED);
        pushcolor(BG_WHITE);
        printf("\nScheme debug:\nInitial id length = %d, ID=",length_read);
        (void)print_ascii(inptr,length_read,note_fileoffset);
        printf(" at file offset %lu\n",note_fileoffset);
    }

    /* Check that the string is printable, and truncate at the first  */
    /* non-printable found                                            */
    id = check_printable(id,MAX_IDLEN);
    id_length = strlen(id);
    if((PRINT_SCHEME_DEBUG) && (id_length != length_read))
        printf("\tchecked id length = %d\n",id_length);
    switch(id_length)
    {
        case -1:    /* read failed                                    */
        case 0:     /* normal or unknown                              */
        case 1:
        case 2:
            /* ###% allow 1 and 2 to be checked?                      */
            if(PRINT_SCHEME_DEBUG)
                printf(" force PLAIN IFD with no offset\n");
            scheme_type = PLAIN_IFD_SCHEME;
            note_byteorder = byteorder;
            break;
        default:
            ifd_offset = 0L;
            start = ((id_length + 3) * 2) / 2;
            maxchk_offset = start + 8;
            /* look for TIFF header first.                            */
            /* At least one maker uses odd offsets, so we can't       */
            /* increment by two                                       */
            for( ; start < maxchk_offset; ++start)
            {
                int print_marker = 0;

                if(PRINT_SCHEME_DEBUG)
                {
                    printf(" header check offset %d",start);
                    print_marker++;
                }

                alt_byteorder = valid_tiffhdr(inptr,note_fileoffset + start,print_marker);
                if(alt_byteorder)
                {
                    scheme_type = HAS_ID_PLUS_TIFFHDR_SCHEME;
                    tiffhdr_offset = start;
                    fileoffset_base = note_fileoffset + tiffhdr_offset;
                    ifd_offset = read_ulong(inptr,alt_byteorder,fileoffset_base + 4);
                    // note_fileoffset += start;
                    note_byteorder = alt_byteorder;
                    if(PRINT_SCHEME_DEBUG)
                        printf(" found TIFF header af file offset %lu\n",fileoffset_base);
                    break;
                }
                else if(PRINT_SCHEME_DEBUG)
                    printf("\t (NO VALID MARKER)\n");
            }

            if(ifd_offset == 0)     /* No TIFF header                 */
            {
                start = ((id_length + 1) * 2) / 2;
                for( ; start < maxchk_offset; ++start)
                {
                    if(PRINT_SCHEME_DEBUG)
                        printf(" entry check offset %d",start);
                    num_entry_offset = note_fileoffset + start;
                    num_entries = read_ushort(inptr,byteorder,num_entry_offset);
                    if(((num_entries > 0) && (maker_size > 0L) &&
                        (num_entries <= (maker_size / 12))) &&
                            (invalid_ifd_entry(inptr,num_entry_offset + 2,
                                                maker_size,num_entries,
                                                    byteorder) == 0))
                    {
                        if(PRINT_SCHEME_DEBUG)
                            printf(" FOUND entry, num_entries=%d\n",num_entries);
                        scheme_type = HAS_ID_SCHEME;
                        ifd_offset = start;
                        note_byteorder = byteorder;
                        break;
                    }
                    else if(PRINT_SCHEME_DEBUG)
                        printf(" , num_entries=%d\t(INVALID)\n",num_entries);
                }
            }

            if(ifd_offset == 0)     /* Try the other byte order       */
            {
                if(byteorder == TIFF_MOTOROLA)
                    alt_byteorder = TIFF_INTEL;
                else
                    alt_byteorder = TIFF_MOTOROLA;
                start = ((id_length + 3) * 2) / 2;
                for( ; start < maxchk_offset; ++start)
                {
                    if(PRINT_SCHEME_DEBUG)
                        printf(" alt byteorder check offset %d",start);
                    num_entry_offset = note_fileoffset + start;
                    num_entries = read_ushort(inptr,alt_byteorder,num_entry_offset);
                    if(((num_entries > 0) && (maker_size > 0L) &&
                        (num_entries <= (maker_size / 12))) &&
                        (invalid_ifd_entry(inptr,num_entry_offset + 2,maker_size,num_entries,alt_byteorder) == 0))
                    {
                        if(PRINT_SCHEME_DEBUG)
                        {
                            printf(" num_entries=%d, VALID entry found, byteorder=",num_entries);
                            print_byteorder(alt_byteorder);
                            putchar('\n');
                        }
                        note_byteorder = alt_byteorder;
                        scheme_type = HAS_ID_SCHEME;
                        ifd_offset = start;
                        break;
                    }
                    else if(PRINT_SCHEME_DEBUG)
                        printf(" , num_entries=%d\t(INVALID)\n",num_entries);
                }
            }
            break;
    }

    if(PRINT_SCHEME_DEBUG)
    {
        printf(" reading num_entries at offset %lu, byteorder=",
                    note_fileoffset + tiffhdr_offset + ifd_offset);
        print_byteorder(note_byteorder);
        putchar('\n');
    }
    if(ifd_offset >= 0)
        num_entries = read_ushort(inptr,note_byteorder,note_fileoffset + tiffhdr_offset + ifd_offset);
    else
        num_entries = 0;    /* read failed */

    if(PRINT_SCHEME_DEBUG)
        printf(" num_entries=%d\n",num_entries);

    
    if(((maker_size > 0L) && (num_entries > (maker_size / 12))) ||
        invalid_ifd_entry(inptr,note_fileoffset + tiffhdr_offset + ifd_offset + 2,
            maker_size,num_entries,note_byteorder))
    {
        /* Initialize the structure                                   */
        maker_scheme.scheme_type = UNKNOWN_SCHEME;
        maker_scheme.note_fileoffset = note_fileoffset;
        maker_scheme.note_length = maker_size;
        maker_scheme.id = id;   /* ...could be one                    */
        maker_scheme.id_length = id_length;
        maker_scheme.tiffhdr_offset = 0L;
        maker_scheme.ifd_offset = 0L;
        maker_scheme.fileoffset_base = fileoffset_base;
        maker_scheme.num_entries = 0;
        maker_scheme.offsets_relative_to = 0;
        maker_scheme.no_nextifd_offset = 1;
        maker_scheme.tiff_byteorder = byteorder;
        maker_scheme.note_byteorder = 0;
        find_maker_model(&maker_scheme,make,model);
        /* If find_maker_model() is changed to succeed at anything,   */
        /* this will have to change...                                */
        if(PRINT_SCHEME_DEBUG)
        {
            pushcolor(GREEN);
            printf(" No valid scheme found ");
            popcolor();
            popcolor();
        }
    }
    else
    {
        if(PRINT_SCHEME_DEBUG)
        {
            pushcolor(GREEN);
            printf(" Found valid scheme ");
            popcolor();
            popcolor();
        }
        /* We seem to have a winner...                                */
        if(scheme_type == UNKNOWN_SCHEME)
            maker_scheme.scheme_type = PLAIN_IFD_SCHEME;
        else
            maker_scheme.scheme_type = scheme_type;
        maker_scheme.note_fileoffset = note_fileoffset;
        maker_scheme.note_length = maker_size;
        maker_scheme.id = id;
        maker_scheme.id_length = id_length;
        maker_scheme.ifd_offset = ifd_offset;
        maker_scheme.tiffhdr_offset = tiffhdr_offset;
        maker_scheme.num_entries = num_entries;
        maker_scheme.no_nextifd_offset = 0;
        maker_scheme.offsets_relative_to = START_OF_TIFFHEADER;
        maker_scheme.fileoffset_base = fileoffset_base;

        /* The two entries allow print_scheme() to make plain when    */
        /* the note byte order gratuitously differs from the rest of  */
        /* the file.                                                  */
        maker_scheme.tiff_byteorder = byteorder;
        maker_scheme.note_byteorder = note_byteorder;

        switch(make)
        {
            case MAKER_ASAHI:
                if(maker_scheme.scheme_type == PLAIN_IFD_SCHEME)
                    maker_scheme.offsets_relative_to = START_OF_ENTRY;
                break;
            case MAKER_FUJIFILM:
            case MAKER_LEICA:
                maker_scheme.offsets_relative_to = START_OF_NOTE;
                maker_scheme.fileoffset_base = note_fileoffset;
                break;
            case MAKER_KYOCERA:
                maker_scheme.offsets_relative_to = START_OF_ENTRY;
                maker_scheme.no_nextifd_offset = 1;
                break;
            case MAKER_PANASONIC:
            case MAKER_TOSHIBA:
                maker_scheme.no_nextifd_offset = 1;
                break;
            case MAKER_SIGMA:
                if(model == SIGMA_SD9)
                    maker_scheme.no_nextifd_offset = 1;
                break;
            default:
                break;
        }
    }
    if(PRINT_SCHEME_DEBUG)
        popcolor();
    return(&maker_scheme);
}

/* Look up a make and model in a table of known sports. This is       */
/* currently just a dummy, but might be used to handle non-ifd notes  */
/* by setting a new type here and revising the generic code to        */
/* dispatch a specialized routine. I have no information on such      */
/* notes now.                                                         */

void
find_maker_model(struct maker_scheme *maker_scheme,int make,int model)
{
        /* do a lookup here on make and model                         */
        /* This should probably return a scheme structure, or some    */
        /* indication of success or failure, but since it's not used  */
        /* now (except to initialize the structure)...                */
}

unsigned short
valid_tiffhdr(FILE *inptr,unsigned long offset,int print_marker)
{
    unsigned short marker = 0;

    if(inptr && offset)
    {
        marker = read_imagemarker(inptr,offset);
        if(print_marker)
            printf(" %#x ",marker);
        switch(marker)
        {
            case TIFF_INTEL:
            case TIFF_MOTOROLA:
                break;
            default:
                marker = 0;
                break;
        }
    }
    return(marker);
}


int
invalid_ifd_entry(FILE *inptr,unsigned long offset,unsigned long segment_size,
                    int nentries,unsigned short byteorder)
{
    struct ifd_entry *entry_ptr;
    int invalid = 0;

    if(inptr && offset && byteorder)
    {
        /* Read an entry at the given offset and see if has a valid   */
        /* type, and non-zero count.                                  */
        entry_ptr = read_entry(inptr,byteorder,offset);
//printf("\n\toffset=%lu type=%u count=%lu, segsize=%lu nentries=%d",offset,entry_ptr->value_type,entry_ptr->count,segment_size,nentries);
        if((entry_ptr->value_type == 0) || (entry_ptr->value_type > DOUBLE) ||
            ferror(inptr) || feof(inptr))
                ++invalid;
        else if(entry_ptr->count == 0)
                ++invalid;
        
        /* If there is more than one entry, and the first appears     */
        /* valid, check the second entry.                             */
        /* If there is only one entry, and it appears valid, make     */
        /* sure that it's size fits within the note. This size check  */
        /* can fail on multi-entry notes when moronic makers write a  */
        /* thumbnail offset, then write the thumbnail outside the     */
        /* note...so we don't check size for multi.                   */
        if(nentries > 1)
        {
            if(invalid == 0)    /* double check                           */
            {
                entry_ptr = read_entry(inptr,byteorder,HERE);
//printf("\n\ttype=%u count=%lu, segsize=%lu\n",entry_ptr->value_type,entry_ptr->count,segment_size);
                if((entry_ptr->value_type == 0) || (entry_ptr->value_type > DOUBLE) ||
                    ferror(inptr) || feof(inptr))
                        ++invalid;
                else if(entry_ptr->count == 0)
                        ++invalid;
            }
        }
        else if(entry_ptr->count > segment_size)
            ++invalid;
    }
    else
        ++invalid;
    return(invalid);
}

int
maker_is_long_offset(struct ifd_entry *entry,int make,int model)
{
    int is_offset = 0;

    if(entry)
    {
        switch(make)
        {
            case MAKER_ASAHI:
                is_offset = maker_asahi_is_long_offset(entry,model);
                break;
            case MAKER_CASIO:
                is_offset = maker_casio_is_long_offset(entry,model);
                break;
            case MAKER_MINOLTA:
                is_offset = maker_minolta_is_long_offset(entry,model);
                break;
            default:
                break;
        }
    }
    return(is_offset);
}

/* Copy the argument string into static storage, stopping at any      */
/* unprintable byte which may be found.                               */
/* Return the null-terminated result string.                          */

char *
check_printable(char *id, int maxlen)
{
    static char buffer[MAX_IDLEN];
    char *idval,*bufval;
    int i = 0;

    memset(buffer,'\0',MAX_IDLEN);
    if(maxlen > MAX_IDLEN)
        maxlen = MAX_IDLEN;
    bufval = buffer;

    for(idval = id; idval && (i < maxlen - 1); ++idval,++bufval,++i)
        if(!isprint(*idval))
        {
            *bufval = '\0';
            break;
        }
        else
            *bufval = *idval;
    return(buffer);
}

/* Provide maker-specific files access to a copy of the scheme. This  */
/* may permit better control over unknown models, e.g. via the Id     */
/* string.  It also provides access to number of entries.             */
struct maker_scheme *
retrieve_makerscheme()
{
    static struct maker_scheme maker_scheme_copy;

    /* structure assignment                                           */
    maker_scheme_copy = maker_scheme;
    return(&maker_scheme_copy);
}
