/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/*          EXIFPROBE - TIFF/JPEG/EXIF image file probe               */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* Copyright (C) 2002 by Duane H. Hesser. All rights reserved.        */
/*                                                                    */
/* See the file LICENSE.EXIFPROBE for terms of use.                   */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#ifndef lint
static char *ModuleId = "@(#) $Id: maker_sanyo.c,v 1.18 2005/01/27 22:13:42 alex Exp $";
#endif

/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* sanyo camera maker-specific routines                             */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* The information coded here is largely due to John Hawkins as shown */
/* at:                                                                */
/*     http://www.exif.org/makernotes/SanyoMakerNote.html             */

/* (John even lists the JPEG thumbnail in TIFF makernotes and the     */
/* Epson PrintIM tag!)                                                */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "extern.h"
#include "misc.h"
#include "maker.h"

/* Find the identifying number assigned to known Sanyo camera models. */
/* This number is used to dispatch print and interpret routines       */
/* approopriate to the current image.                                 */

int
sanyo_model_number(char *model,char *software)
{
    int number = GENERIC;

    if(model)
    {
        switch(*model)
        {
            case 'S':
                if(strcmp(model,"SR6 ") == 0)
                    number = SANYO_SR6;
                else if(strcmp(model,"SX113") == 0)
                    number = SANYO_SX113;
                break;
            default:
                number = GENERIC;
                break;
        }
    }
    return(number);
}

/* Dispatch a print routine based upon model                          */

void
print_sanyo_makervalue(struct ifd_entry *entry_ptr,int make,int model,
                    char *prefix)
{

    if(entry_ptr && (PRINT_VALUE))
    {
        switch(model)
        {
            case SANYO_SR6:
            case SANYO_SX113:
            case GENERIC:
            default:
                print_sanyo1_makervalue(entry_ptr,make,model,prefix);
                sanyo1_interpret_value(entry_ptr);
                break;
        }
    }
}

/* Model-specific print routine for sanyo cameras. This routine is      */
/* responsible for picking off any direct entry tags which are        */
/* peculiar and will not be handled properly by print_value()         */
/* (usually UNDEFINED values which fit in the 4-byte entry value). If */
/* there are no such entries, this function simply calls              */
/* print_value().                                                     */

void
print_sanyo1_makervalue(struct ifd_entry *entry_ptr,int make,int model,
                    char *prefix)
{
    if(entry_ptr && (PRINT_VALUE))
    {
        switch(entry_ptr->tag) 
        {
            default:
                print_value(entry_ptr,prefix);
                break;
        }
    }
}


/* Dispatch a routine to decode and print offset values for sanyo     */
/* cameras.                                                           */
void
print_sanyo_offset_makervalue(FILE *inptr,unsigned short byteorder,
    struct ifd_entry *entry_ptr,unsigned long offset_ref,char *prefix,
    int indent,int addrwidth,int make,int model,int at_offset)
{

    if(entry_ptr && (PRINT_VALUE))
    {
        switch(model)
        {
            default:
                sanyo1_offset_makervalue(inptr,byteorder,entry_ptr,
                                            offset_ref,prefix,indent,
                                            addrwidth,make,model,at_offset);
                sanyo1_interpret_offset_makervalue(inptr,byteorder,entry_ptr,
                                            offset_ref);
                break;
        }
    }
}

/* Model-specific routine to print UNDEFINED values found at offsets  */
/* in sanyo makernotes. One of these may be supplied for each model   */
/* with unique UNDEFINED tags.                                        */

void
sanyo1_offset_makervalue(FILE *inptr,unsigned short byteorder,
    struct ifd_entry *entry_ptr,unsigned long offset_ref,char *prefix,
    int indent,int addrwidth,int make,int model,int at_offset)
{
    unsigned long value_offset;
    unsigned long dumplength;
    unsigned short marker;
    char *nameoftag;

    if(entry_ptr)
    {
        nameoftag = maker_tagname(entry_ptr->tag,make,model); 
        value_offset = offset_ref + entry_ptr->value;
        /* It's not clear that PrintIM needs a separate case (the     */
        /* default would do), except that some day maybe we'll know   */
        /* how to interpret it.                                       */
        switch(entry_ptr->tag)
        {
            case 0x0100:    /* Jpeg Thumbnail in TIFF MakerNote       */
                if(at_offset)
                {
                    print_tag_address(VALUE_AT_OFFSET,value_offset,indent,prefix,addrwidth);
                    extraindent(SUBINDENT);
                    printf("Start of JPEG Thumbnail");
                }
                if(!(PRINT_OFFSET))
                {
                    if(PRINT_BOTH_VALUE)
                        printf("@%#lx=%lu",value_offset,value_offset);
                    else if(PRINT_HEX_VALUE)
                        printf("@%#lx",value_offset);
                    else if(PRINT_DEC_VALUE)
                        printf("@%lu",value_offset);
                }
                printf(" length %ld",entry_ptr->count);
                marker = read_imagemarker(inptr,value_offset);
                if(PRINT_SEGMENT)
                {
                    putchar('\n');
                    (void)process_jpeg_segments(inptr,value_offset,
                                                marker,entry_ptr->count,
                                                indent+(SUBINDENT * 2),addrwidth);
                    print_tag_address(SEGMENT,value_offset + entry_ptr->count - 1,
                                                indent,"-",addrwidth);
                    extraindent(SUBINDENT);
                    printf("End of JPEG Thumbnail");
                }
                /* make certain we're at the end unless 'inptr' bad   */
                fseek(inptr,value_offset + entry_ptr->count,0);
                break;
            case 0x0e00: /* PrintIM (Epson Print Image matching)      */
                if(!(PRINT_OFFSET))
                {
                    if(PRINT_BOTH_VALUE)
                        printf("@%#lx=%lu\n",value_offset,value_offset);
                    else if(PRINT_HEX_VALUE)
                        printf("@%#lx\n",value_offset);
                    else if(PRINT_DEC_VALUE)
                        printf("@%lu\n",value_offset);
                }
                process_pim(inptr,byteorder,entry_ptr->value,offset_ref,
                    entry_ptr->count,nameoftag,"Exif_Maker_",prefix,indent,addrwidth);
                break;
            case 0x0f00: /* Data                                      */
                if(at_offset)
                {
                    print_tag_address(SEGMENT,value_offset,indent,prefix,addrwidth);
                    print_makertagid(entry_ptr,19," : ",SUBINDENT,make,model);
                }
                else if(PRINT_OFFSET)
                        printf(" = ");
                printf("UNDEFINED, @%lu count %lu",value_offset, entry_ptr->count);
                if(Max_undefined == 0)
                    printred(" (not dumped, use -U)");
                else
                {
                    if((Max_undefined == DUMPALL)
                        || (Max_undefined > entry_ptr->count))
                            dumplength = entry_ptr->count;
                    else 
                        dumplength = Max_undefined;
                    putchar('\n');
                    hexdump(inptr,value_offset,entry_ptr->count,dumplength,12,
                                indent,SUBINDENT,addrwidth);
                }
                /* make certain we're at the end unless 'inptr' bad   */
                fseek(inptr,value_offset + entry_ptr->count,0);
                break;
            default:
                print_generic_offset_makervalue(inptr,byteorder,entry_ptr,
                                        offset_ref,prefix,indent,
                                        addrwidth,make,model,at_offset);
                break;
        }
    }
}


/* sanyo-specific tagnames for makernotes.                            */

char *
maker_sanyo_tagname(unsigned short tag,int model)
{
    char *tagname = (char *)0;

    /* Check specific models first                                    */
    switch(model)
    {
        /* Add special-case models here                               */
        default:
            break;
    }

    /* If no model-specific tag is found, use a "generic" Sanyo tag   */
    if(tagname == NULL)
    {
        switch(tag)
        {
            case 0x0100: tagname = "JpegThumbnail"; break;
            case 0x0200: tagname = "SpecialMode"; break;
            case 0x0201: tagname = "Quality"; break;
            case 0x0202: tagname = "Macro"; break;
            case 0x0204: tagname = "DigiZoom"; break;
            case 0x0207: tagname = "SoftwareRelease"; break;
            case 0x0208: tagname = "PictureInfo"; break;
            case 0x0209: tagname = "CameraID"; break;
            case 0x020e: tagname = "SeqShotMethod"; break;
            case 0x020f: tagname = "WideRange"; break;
            case 0x0210: tagname = "ColorAdjMode"; break;
            case 0x0213: tagname = "QuickShot"; break;
            case 0x0214: tagname = "SelfTImer"; break;
            case 0x0216: tagname = "VoiceMemo"; break;
            case 0x0217: tagname = "RecordShutterRel"; break;
            case 0x0218: tagname = "FlickerReduce"; break;
            case 0x0219: tagname = "OpticalZoom"; break;
            case 0x021b: tagname = "DigitalZoom"; break;
            case 0x021d: tagname = "LightSourceSpecial"; break;
            case 0x021e: tagname = "Resaved"; break;
            case 0x021f: tagname = "SceneSelect"; break;
            case 0x0223: tagname = "ManualFocalDist"; break;
            case 0x0224: tagname = "SeqShotIntvl"; break;
            case 0x0225: tagname = "FlashMode"; break;
            case 0x0e00: tagname = "PrintIM"; break;
            case 0x0f00: tagname = "Data"; break;
            default: break;
        }
    }
    return(tagname);
}


void
sanyo1_interpret_value(struct ifd_entry *entry_ptr)
{
    switch(entry_ptr->tag)
    {
        case 0x0201:    /* Quality */
            print_startvalue();
            /* ###%%% is this related to image size, e.g. 640x480?    */
            /* This tag is the same tag number as used by the         */
            /* Olympus, Epson, and Minolta F100, except that "super   */
            /* high" + "super fine" would give a (short) value of 34, */
            /* which is the value used for uncompressed tiff in the   */
            /* Epson at 2048x1536.                                    */
            switch(entry_ptr->value & 0xff)
            {
                case 0: printf("very low res,"); break;
                case 1: printf("low res,"); break;
                case 2: printf("medium low res,");    /* 640 x 480 */ break;
                case 3: printf("medium res,"); break;
                case 4: printf("medium high res,"); break;
                case 5: printf("high res,"); break;
                case 6: printf("very high res,"); break;
                case 7: printf("super high"); break;
                default: printred("undefined"); break;
            }
            /* ###%%% is this related to compression, e.g TIFF vs     */
            /* JPEG ???                                               */
            switch((entry_ptr->value >> 8) & 0xff)
            {
                case 0: printf(" normal"); break;
                case 1: printf(" fine"); break;
                case 2: printf(" super fine"); break;
            }
            print_endvalue();
            break;
        case 0x0202:    /* Macro */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Normal"); break;
                case 1: printf("Macro"); break;
                case 2: printf("View"); break;
                case 3: printf("Manual"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x020e:    /* Sequential Shot Method */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("none"); break;
                case 1: printf("Standard"); break;
                case 2: printf("Best"); break;
                case 3: printf("Exposure Adjust"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x020f:    /* Wide Range */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Off"); break;
                case 1: printf("On"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x0210:    /* Color Adjustment Mode */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Off"); break;
                case 1: printf("On"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x0213:    /* Quick Shot */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Off"); break;
                case 1: printf("On"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x0214:    /* Self Timer  */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Off"); break;
                case 1: printf("On"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x0216:    /* Voice Memo  */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Off"); break;
                case 1: printf("On"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x0217:    /* Record Shutter Release  */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Held"); break;
                case 1: printf("Start/Stop"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x0218:    /* Flicker Reduce  */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Off"); break;
                case 1: printf("On"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x0219:    /* Optical Zoom  */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Disabled"); break;
                case 1: printf("Enabled"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x021b:    /* Digital Zoom  */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Disabled"); break;
                case 1: printf("Enabled"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x021d:    /* Light Source Special  */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Off"); break;
                case 1: printf("On"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x021e:    /* Resaved  */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("No"); break;
                case 1: printf("Yes"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x021f:    /* Scene Select  */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Off"); break;
                case 1: printf("Sport"); break;
                case 2: printf("TV?"); break;
                case 3: printf("Night"); break;
                case 4: printf("User 1"); break;
                case 5: printf("User 2"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x0224:    /* Sequential Shot Interval  */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("5 frames/sec"); break;
                case 1: printf("10 frames/sec"); break;
                case 2: printf("15 frames/sec?"); break;
                case 3: printf("20 frames/sec"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x0225:    /* Flash Mode  */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Auto"); break;
                case 1: printf("Force"); break;
                case 2: printf("Disabled"); break;
                case 3: printf("Redeye"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;

        default:
            break;
    }
}

void
sanyo1_interpret_offset_makervalue(FILE *inptr,unsigned short byteorder,
                    struct ifd_entry *entry_ptr,unsigned long offset_ref)
{
    unsigned long offset,value;

    if(entry_ptr)
    {
        offset = entry_ptr->value + offset_ref;
        switch(entry_ptr->tag)
        {
            case 0x0200:    /* SpecialMode                               */
                value = read_ulong(inptr,byteorder,offset);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Normal"); break;
                    case 1: printf("Unkown"); break;
                    case 2: printf("Fast"); break;
                    case 3: printf("Panorama,");
                            value = read_ulong(inptr,byteorder,HERE);
                            printf("#%lu,",value);
                            value = read_ulong(inptr,byteorder,HERE);
                            switch(value)
                            {
                                case 1: printf(" Left to Right"); break;
                                case 2: printf(" Right to Left"); break;
                                case 3: printf(" Bottom to Top"); break;
                                case 4: printf(" Top to Bottom"); break;
                                default: printred(" undefined"); break;
                            }
                            break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            default:
                break;
        }
    }
}
