/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/*          EXIFPROBE - TIFF/JPEG/EXIF image file probe               */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* Copyright (C) 2002 by Duane H. Hesser. All rights reserved.        */
/*                                                                    */
/* See the file LICENSE.EXIFPROBE for terms of use.                   */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#ifndef WALL
static char *ModuleId = "@(#) $Id: maker_sigma.c,v 1.1 2005/01/27 19:35:19 alex Exp $";
#endif

/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* Sigma camera maker-specific routines.                              */
/* Tag information from;                                                  */
/*   http://www.x3f.info/technotes/FileDocs/MakerNoteDoc.html         */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "extern.h"
#include "misc.h"
#include "maker.h"

/* Find the id number assigned to known Sigma camera models.          */
/* This number is used to dispatch print and interpret routines       */
/* apporopriate to the current image.                                 */

int
sigma_model_number(char *model,char *software)
{
	int number = GENERIC;

	switch(*model)
	{
        case 'S':
            if(strcmp(model,"SIGMA SD9") == 0)
                number = SIGMA_SD9;
            else if(strcmp(model,"SIGMA SD10") == 0)
                number = SIGMA_SD10;
            break;
		default:
			number = GENERIC;
			break;
	}
	return(number);
}

/* Dispatch a print routine based upon model                          */

void
print_sigma_makervalue(struct ifd_entry *entry_ptr,int make, int model,
					char *prefix)
{

	switch(model)
	{
		case SIGMA_SD9:
		case SIGMA_SD10:
		case GENERIC:
		default:
			print_sigma1_makervalue(entry_ptr,make,model,prefix);
			break;
	}
}

/* Model-specific print routine for sigma cameras. This routine is    */
/* responsible for picking off any direct entry tags which are        */
/* peculiar and will not be handled properly by print_value()         */
/* (usually UNDEFINED values which fit in the 4-byte entry value). If */
/* there are no such entries, this function simply calls              */
/* print_value().                                                     */

void
print_sigma1_makervalue(struct ifd_entry *entry_ptr,int make, int model,
					char *prefix)
{
	if(entry_ptr)
	{
		switch(entry_ptr->tag) 
		{
			/* pick off values that need special handling in cases    */
            /* here                                                   */
			default:
				print_value(entry_ptr,PREFIX);
				break;
		}
	}
}

/* Dispatch a routine to decode and print offset values for sigma     */
/* cameras.                                                           */

void
print_sigma_offset_makervalue(FILE *inptr,unsigned short byteorder,
	struct ifd_entry *entry_ptr,unsigned long offset_ref,char *prefix,
	int indent,int addrwidth,int make,int model,int at_offset)
{

	switch(model)
	{
		case SIGMA_SD9:
		case SIGMA_SD10:
		default:
			sigma1_offset_makervalue(inptr,byteorder,entry_ptr,
										offset_ref,prefix,indent,
										addrwidth,make,model,at_offset);
			break;
	}
}

/* Model-specific routine to print values found at an offset from the */
/* IFD entry in sigma makernotes.                                     */

void
sigma1_offset_makervalue(FILE *inptr,unsigned short byteorder,
	struct ifd_entry *entry_ptr,unsigned long offset_ref,char *prefix,
	int indent,int addrwidth,int make,int model,int at_offset)
{
	unsigned long value_offset;
	unsigned long dumplength;
	unsigned short marker;
	char *nameoftag;

	if(entry_ptr)
	{
		nameoftag = maker_tagname(entry_ptr->tag,make,model); 

		switch(entry_ptr->tag)
		{
			default:
				print_generic_offset_makervalue(inptr,byteorder,entry_ptr,
										offset_ref,prefix,indent,
										addrwidth,make,model,at_offset);
				break;
		}
	}
}

char *
maker_sigma_tagname(unsigned short tag,int model)
{
	char *tagname = (char *)0;

	/* Check specific models first                                    */
	switch(model)
	{
		case SIGMA_SD9:
		case SIGMA_SD10:
		default:
			tagname = maker_sigma1_tagname(tag,model);
			break;
	}
	return(tagname);
}

/* Tags for the older cameras.                                        */

char *
maker_sigma1_tagname(unsigned short tag,int model)
{
	char *tagname = (char *)0;

	switch(tag)
	{
        case 0x0002: tagname = "CameraSerialNumber"; break;
        case 0x0003: tagname = "DriveMode"; break;
        case 0x0004: tagname = "ResolutionMode"; break;
        case 0x0005: tagname = "AutofocusMode"; break;
        case 0x0006: tagname = "FocusSetting"; break;
        case 0x0007: tagname = "WhiteBalance"; break;
        case 0x0008: tagname = "ExposureMode"; break;
        case 0x0009: tagname = "MeteringMode"; break;
        case 0x000A: tagname = "LensFocalLengthRange"; break;
        case 0x000B: tagname = "ColorSpace"; break;
        case 0x000C: tagname = "Exposure"; break;
        case 0x000D: tagname = "Contrast"; break;
        case 0x000E: tagname = "Shadow"; break;
        case 0x000F: tagname = "Highlight"; break;
        case 0x0010: tagname = "Saturation"; break;
        case 0x0011: tagname = "Sharpness"; break;
        case 0x0012: tagname = "X3FillLight"; break;
        case 0x0014: tagname = "ColorAdjustment"; break;
        case 0x0015: tagname = "AdjustmentMode"; break;
        case 0x0016: tagname = "Quality"; break;
        case 0x0017: tagname = "Firmware"; break;
        case 0x0018: tagname = "Software"; break;
        case 0x0019: tagname = "AutoBracket"; break;
			default: break;
	}
	return(tagname);
}

